# -*- coding: utf-8 -*-

# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# Contributors: Gauthier Bastien, Fabio Marcuzzi, IMIO
# ------------------------------------------------------------------------------
import re, xml.sax, math

from appy.pod import *
from appy.pod.odf_parser import OdfEnvironment
from appy.shared.css import CssStyles, CssValue, px2cm
from appy.shared.xml_parser import XmlEnvironment, XmlParser, Escape
from appy.shared.utils import WhitespaceCruncher, formatNumber, addPair
from appy.pod.styles_manager import Style,BulletedProperties,NumberedProperties

# Tags for which there is a direct correspondance between HTML and ODF
h = 'text:h'
p = 'text:p'
span = 'text:span'
tlist = 'text:list'
table = 'table:table'
cell = table + '-cell'
HTML_2_ODF = {
  'h1':h, 'h2':h, 'h3':h, 'h4':h, 'h5':h, 'h6':h, 'p':p, 'div':p,
  'blockquote':p, 'address':p, 'sub': span, 'sup': span, 'br':'text:line-break',
  'table': table, 'thead': table + '-header-rows',
  'tr': table + '-row', 'td': cell, 'th': cell, 'a': 'text:a',
  'ol': tlist, 'ul': tlist, 'li': 'text:list-item'}

# Inner tags whose presence is only useful for specifying style information
STYLE_ONLY_TAGS = ('b', 'strong', 'i', 'em', 'strike', 's', 'u', 'span', 'q',
                   'code', 'font', 'samp', 'kbd', 'var', 'label', 'abbr')
for tag in STYLE_ONLY_TAGS: HTML_2_ODF[tag] = 'text:span'

# Styles whose translation to ODF is simple
SIMPLE_TAGS = ('h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'div', 'blockquote',
               'address', 'sub', 'sup', 'br', 'th', 'td')

INNER_TAGS_NO_BR = STYLE_ONLY_TAGS + ('sub','sup','a','acronym','abbr','img')
INNER_TAGS = INNER_TAGS_NO_BR + ('br',)
TABLE_CELL_TAGS = ('td', 'th')
TABLE_COL_TAGS = TABLE_CELL_TAGS + ('col',)
TABLE_ROW_TAGS = ('tr', 'colgroup')
OUTER_TAGS = TABLE_CELL_TAGS + ('li',)
PARA_TAGS = ('p', 'div', 'blockquote', 'address')
# The following elements can't be rendered inside paragraphs
NOT_INSIDE_P = XHTML_HEADINGS + XHTML_LISTS + ('table',)
NOT_INSIDE_P_OR_P = NOT_INSIDE_P + PARA_TAGS
NOT_INSIDE_LI =  XHTML_HEADINGS + ('table',) + PARA_TAGS
NOT_INSIDE_LIST = ('table',)
IGNORABLE_TAGS = ('meta', 'title', 'style', 'script')

# ------------------------------------------------------------------------------
class Element:
    '''Abstract base class representing any HTML element encountered during the
       SAX parsing.'''

    # An instance of one of its concrete classes (see below) will be pushed on
    # the stack of currently parsed elements every time such element is
    # encountered.

    def getClass(self, last=False):
        '''Returns the name of the CSS class(es) defined on p_self'''
        styles = self.cssStyles
        if not styles: return
        return styles.getClass(last=last)

    def hasClass(self, css):
        '''Has p_self a CSS class named p_css ?'''
        styles = self.cssStyles
        return styles and styles.hasClass(css) or False

    def noClass(self):
        '''Returns True if no CSS class is defined on p_self'''
        return not self.getClass()

# ------------------------------------------------------------------------------
class HtmlElement(Element):
    '''Represents any HTML element'''

    # Moreover, if the element is a table, it will be wrapped in a HtmlTable
    # element (see below).

    elemTypes = {'p':'para', 'div':'para', 'blockquote': 'para',
                 'li':'para', 'ol':'list', 'ul':'list'}

    # Prototypical instances
    protos = {}

    def __init__(self, elem, attrs=None, parent=None):
        self.elem = elem
        # Store a reference to the parent element if known
        self.parent = parent
        # Extract style information from p_elem and p_attrs
        self.cssStyles = CssStyles(elem, attrs)
        # When merging styles, "cssStyles" may temporarily hold a set of merged
        # styles. In this case, the original CssStyles instance is stored in the
        # following attribute.
        self.originalStyles = None
        # Extract other attributes depending on p_elem
        self.attrs = self.extractAttributes(elem, attrs)
        # When the HTML element corresponding to p_self is completely dumped,
        # if there was a problem related to tags inclusion, we may need to dump
        # start tags corresponding to tags that we had to close before dumping
        # this element.
        self.tagsToReopen = None # ~[HtmlElement]~
        # Before dumping the closing tag corresponding to p_self, we may need to
        # close other tags (ie closing a paragraph before closing a cell).
        self.tagsToClose = None # ~[HtmlElement]~
        # This tag's type
        if self.elemTypes.has_key(elem):
            self.elemType = self.elemTypes[elem]
        else:
            self.elemType = elem
        # If a conflict occurs on this element, we will note it
        self.isConflictual = False
        # Sometimes we must remember the ODT style that has been computed and
        # applied to this HTML element (for lists).
        self.odStyle = None
        # For cells, the style to apply to their inner paragraphs (set later)
        self.innerStyle = None
        # We must know if a style must be applied on inner paragraphs within
        # "li" tags.
        if (elem == 'li') and attrs.has_key('class'):
            self.paraStyle = attrs['class']
        else:
            self.paraStyle = None # Will also be set on "ol" and "ul" tags.
        # Must we only keep tag content and remove the tag itself ?
        self.removeTag = False
        # Must we dump a line-break just before encountering a tag from
        # INNER_TAGS within this tag ?
        self.dumpLineBreakOnInner = False
        # Was content already dumped from this tag (or any of its sub-tags) into
        # the result ?
        self.contentDumped = False
        # "Dump status" can have 3 values:
        # - None       in the majority of cases;
        # - "waiting"  we don't know yet if we must dump this tag or not: it
        #              will depend if content will be found or not;
        # - "dumped"   a tag previously "waiting" has effectively be dumped.
        self.dumpStatus = None
        # Inner CSS styles are those from a removed inner tag, or from this tag,
        # being an outer tag whose CSS styles must be dumped to an inner tag
        # that will be added.
        self.innerCssStyles = None
        if elem == 'li': self.addInnerCssStyles(self)
        # The number (starting at 1) of p_self's currently walked sub-tag. If we
        # are not walking a sub-tag, this number is set to 0.
        self.subNumber = 0

    def extractAttributes(self, elem, attrs):
        '''Extracts useful attributes from p_attrs'''
        if not attrs: return
        if elem == 'ol':
            if attrs.has_key('start'):
                return {'start': int(attrs['start'])}
        elif elem in TABLE_COL_TAGS:
            return attrs

    def getOdfTag(self):
        '''Gets the raw ODF tag that corresponds to me'''
        return HTML_2_ODF.get(self.elem, '')

    def getOdfTags(self):
        '''Gets the start and end tags corresponding to p_self'''
        tag = self.getOdfTag()
        if not tag: return (None, None)
        return ('<%s>' % tag, '</%s>' % tag)

    def setConflictual(self):
        '''Note p_self as conflictual'''
        self.isConflictual = True
        return self

    def getPath(self):
        '''Return this element's "absolute path" within the XHTML tree'''
        res = self.elem
        if self.parent:
            res = '%s>%s' % (self.parent.getPath(), res)
        return res

    def getConflictualElements(self, env):
        '''p_self was just parsed. In some cases, this element can't be dumped
           in the result because there are conflictual elements among previously
           parsed opening elements (p_env.currentElements). For example, if we
           just dumped a "p", we can't dump a table within the "p". Such
           constraints do not hold in XHTML code but hold in ODF code.'''
        if not env.currentElements: return ()
        parent = env.getCurrentElement()
        # Special case: check elements that can't be found within a "li".
        # Instead of being noted as "conflictual", note that we must keep
        # these element's contents but remove the surrounding tags.
        if (parent.elem == 'li') and (self.elem in PARA_TAGS):
            self.removeTag = True
            parent.addInnerCssStyles(self)
            return ()
        # Check elements that can't be found within a paragraph / li. The list
        # of such elements is different for a "li", but "li" has elemType
        # "para", so we must explicitly check the "li" case before checking
        # "elemType" being "para" or not.
        if parent.elem == 'li':
            if self.elem in NOT_INSIDE_LI:
                return (parent.parent.setConflictual(),parent.setConflictual(),)
        elif (parent.elemType == 'para') and (self.elem in NOT_INSIDE_P_OR_P):
            return (parent.setConflictual(),)
        # Check inner paragraphs
        elif (parent.elem in INNER_TAGS) and (self.elemType == 'para'):
            res = [parent.setConflictual()]
            if len(env.currentElements) > 1:
                i = 2
                visitParents = True
                while visitParents:
                    try:
                        nextParent = env.currentElements[-i]
                        i += 1
                        res.insert(0, nextParent.setConflictual())
                        if nextParent.elemType == 'para':
                            visitParents = False
                    except IndexError:
                        visitParents = False
            return res
        if parent.tagsToClose and \
            (parent.tagsToClose[-1].elemType == 'para') and \
            (self.elem in NOT_INSIDE_P_OR_P):
            return (parent.tagsToClose[-1].setConflictual(),)
        # Check elements that can't be found within a list
        if (parent.elemType == 'list') and (self.elem in NOT_INSIDE_LIST):
            return (parent.setConflictual(),)
        return ()

    def addInnerParagraph(self, env):
        '''Dump an inner paragraph inside self (if not already done)'''
        if self.tagsToClose: return # We already did it
        dump = env.dumpString
        p = HtmlElement('p', parent=self)
        dump('<text:p')
        if self.elem == 'li':
            listElem = env.getCurrentElement(isList=True)
            itemStyle = listElem.elem # ul or ol
            # Which 'li'-related style must I use ?
            css = self.cssStyles.classes
            if css:
                odtStyle = env.findStyle(self)
                styleName = None
                if odtStyle:
                    if odtStyle.name == 'podItemKeepWithNext':
                        itemStyle += '_kwn'
                    else:
                        if css == 'podItemKeepWithNext':
                            # We ignore the odtStyle. To improve: merge
                            # keep-with-next functionality with this style.
                            styleName = env.itemStyles[itemStyle + '_kwn']
                        else:
                            styleName = css
                styleName = styleName or env.itemStyles[itemStyle]
                dump(' text:style-name="%s"' % styleName)
            else:
                # Check if a style must be applied on 'p' tags
                if self.innerCssStyles:
                    p.cssStyles.merge(self.innerCssStyles)
                innerStyle = listElem.paraStyle or env.itemStyles[itemStyle]
                if innerStyle:
                    p.cssStyles.addClass(innerStyle)
                dump(env.getOdfAttributes(p))
        else:
            dump(env.getOdfAttributes(p))
        dump('>')
        if not self.tagsToClose: self.tagsToClose = []
        self.tagsToClose.append(p)
        # Note that an inner paragraph has been added
        self.show(env, prefix='!')

    def addInnerCssStyles(self, xhtmlElem):
        '''Remember on p_self the cssStyles defined on p_xhtmlElem'''
        # Do nothing if inner styles are already there or if p_xhtmlElem has no
        # CSS styles.
        if self.innerCssStyles or not xhtmlElem.cssStyles: return
        self.innerCssStyles = xhtmlElem.cssStyles

    def dump(self, start, env):
        '''Dumps the start or end (depending on p_start) tag of this HTML
           element. We must take care of potential innerTags.'''
        # Compute the tag in itself
        tag = ''
        prefix = '<'
        if not start: prefix += '/'
        # Compute tag attributes
        attrs = ''
        if start:
            if self.elemType == 'list':
                # I must specify the list style
                attrs += ' text:style-name="%s"' % self.odStyle
                if self.elem == 'ol':
                    # I have interrupted a numbered list. I need to continue
                    # the numbering.
                    attrs += ' text:continue-numbering="true"'
            else:
                attrs = env.getOdfAttributes(self)
        tag = prefix + self.getOdfTag() + attrs + '>'
        # Close/open subTags if any
        toClose = self.tagsToClose
        if toClose:
            for subElem in toClose:
                subTag = subElem.dump(start, env)
                if start: tag += subTag
                else: tag = subTag + tag
        return tag

    def getLevel(self):
        '''Gets the "depth" of this tag among the currently walked tags'''
        if not self.parent: return 0
        return self.parent.getLevel() + 1

    def getName(self):
        '''Get the name of this tag, potentially augmented wit sub-tags from
           self.tagsToClose.'''
        res = self.elem
        if self.tagsToClose:
            for tag in self.tagsToClose:
                res += '+' + tag.getName()
        return res

    def inInnerTag(self):
        '''Is this tag in an inner tag ?'''
        parent = self.parent
        if not parent: return
        return (parent.elem in INNER_TAGS) or parent.inInnerTag()

    def __repr__(self, prefix='<'):
        res = self.getName()
        if self.removeTag:
            # We do not dump the prefix, used for indicating the fact that the
            # tag is opened or closed.
            res += '-'
            prefix = ''
        if self.isConflictual: res += '*'
        return '%s%s' % (prefix, res)

    def show(self, env, indented=True, prefix='<', content=None):
        '''Returns a (possibly) p_indented representation of this HTML element
           (or of some other p_content at this indentation level), for debugging
           purposes.'''
        if not env.verbose: return
        if content:
            res = '%s%s' % (prefix, content)
        else:
            res = self.__repr__(prefix=prefix)
        if indented:
            res = (' ' * 2 * self.getLevel()) + res
        print res

# Create prototypical instances
for tag in ('p', 'ul', 'ol'): HtmlElement.protos[tag] = HtmlElement(tag)

# ------------------------------------------------------------------------------
class HtmlTable(Element):
    '''Represents an HTML table, and also a sub-buffer. When parsing elements
       corresponding to an HTML table (<table>, <tr>, <td>, etc), we can't dump
       corresponding ODF elements directly into the global result buffer
       (XhtmlEnvironment.res). Indeed, when dumping an ODF table, we must
       dump columns declarations at the beginning of the table. So before
       dumping rows and cells, we must know how much columns will be present
       in the table. It means that we must first parse the first <tr> entirely
       in order to know how much columns are present in the HTML table before
       dumping the ODF table. So we use this class as a sub-buffer that will
       be constructed as we parse the HTML table; when encountering the end
       of the HTML table, we will dump the result of this sub-buffer into
       the parent buffer, which may be the global buffer or another table
       buffer.'''

    def __init__(self, env, xhtmlElem, attrs):
        self.env = env
        # p_xhtmlElem is the HtmlElement representing this table
        self.parent = xhtmlElem.parent
        self.elem = 'table'
        self.cssStyles = CssStyles(self.elem, attrs)
        self.originalStyles = None
        self.removeTag = False
        self.contentDumped = False
        self.name = env.getUniqueStyleName('table')
        # Suffix the table name with a post-processor command when needed
        if attrs and attrs.has_key('keeprows'):
            # This command tells the post-processor to remove some rows from
            # the table.
            self.name += '(%s)' % attrs['keeprows']
            # So ensure the post-processor is enabled
            env.enablePpp()
        self.styleNs = env.ns[OdfEnvironment.NS_STYLE]
        self.tableNs = env.ns[OdfEnvironment.NS_TABLE]
        # Must be set a border for this table ?
        self.border = self.hasBorder()
        # If a CSS property "border-spacing" is defined on this table,
        # m_setTableStyle will parse it and store it here.
        self.borderSpacing = None
        # Get the TableProperties instance. There is always one.
        self.props = props = env.findStyle(self)
        self.style, self.widthInPx, self.originalWidthInPx =self.setTableStyle()
        # Patch the table name when columns must be modified by LO
        modifier = props.columnModifier
        if modifier:
            prefix = props.columnModifiersPrefixes[modifier]
            self.name = '%s_%s' % (prefix, self.name)
        self.res = u'' # The sub-buffer
        # The temporary sub-buffer, into which we will dump all table
        # sub-elements, until we encounter the end of the first row. Then, we
        # will know how much columns are defined in the table; we will dump
        # columns declarations into self.res and dump self.tempRes into
        # self.res.
        self.tempRes = u''
        # Was the first table row completely parsed ?
        self.firstRowParsed = False
        # The number of columns in the table
        self.nbOfColumns = 0
        # Are we currently within a table cell? Instead of a boolean, the field
        # stores an integer. The integer is > 1 if the cell spans more than one
        # column.
        self.inCell = 0
        # The index, within the current row, of the current cell
        self.cellIndex = -1
        # The size of the content of the currently parsed table cell
        self.cellContentSize = 0
        # The size of the longest word of the currently parsed table cell
        self.cellLongestWord = 0
        # The following lists store, for every column: the size of (a) the
        # longest content and (b) the longest words of all its cells.
        self.columnContentSizes = []
        self.columnLongestWords = [] # Computed but currently not used
        # The following list stores, for every column, its width, if specified.
        # If widths are found, self.columnContentSizes will not be used:
        # self.columnWidths will be used instead.
        self.columnWidths = []

    def hasBorder(self):
        '''Support for table borders is currently limited. We have 2 cases:
           either we set a border to the table, either we do not. This method
           r_eturns a boolean value reflecting this, based on the CSS or HTML
           attribute "border". By default, we set a border.'''
        prop = getattr(self.cssStyles, 'border', None)
        if prop is None: return True # By default we set a border
        val = prop.value
        if (val == '0') or ('none' in val) or ('undefined' in val): return
        return True

    def setTableStyle(self):
        '''The default ODT style "podTable" will apply to the table, excepted if
           specific style elements are specified. In this case, we will create a
           dynamic style whose parent will be "podTable". This method returns a
           tuple (styleName, tableWidthPx). The table width in pixels is
           sometimes needed to convert column widths, expressed in pixels, to
           percentages.'''
        tableProps = self.props
        cssStyles = self.cssStyles
        if hasattr(cssStyles, 'borderspacing'):
            # This attribute will be used at the cell level. Extract it and
            # delete it.
            self.borderSpacing = cssStyles.borderspacing
            del cssStyles.borderspacing
        # Get the table width and alignment
        hasWidth = hasattr(cssStyles, 'width')
        width = tableProps.getWidth(cssStyles)
        originalWidth = tableProps.getWidth(cssStyles, original=True)
        align = getattr(cssStyles, 'textalign', 'left')
        # Get the page width, in cm, and the ratio "px2cm"
        pageWidth, px2cmRatio = tableProps.pageWidth, tableProps.px2cm
        pageWidth = pageWidth or self.env.pageWidth
        px2cmRatio = px2cmRatio or px2cm
        # Compute the table attributes for setting its width
        s = self.styleNs
        if width.unit == '%':
            tableWidth = pageWidth * (width.value / 100.0)
            percentage = str(width.value)
        else: # cm or px
            ratio = (width.unit == 'cm') and 1.0 or px2cmRatio
            tableWidth = min(width.value / ratio, pageWidth)
            percentage = formatNumber(float(tableWidth/pageWidth)*100, sep='.')
        # Compute the table size in PX: it will be needed to convert column
        # widths in px to percentages.
        tableWidthPx = int(tableWidth * px2cmRatio)
        # Get the original table size in PX, that can be different from the real
        # one as computed above (see styles_manager::TableProperties.wideAbove).
        if originalWidth.unit == 'px':
            originalTableWidthPx = int(originalWidth.value)
        else:
            originalTableWidthPx = None
        # Get margins if defined
        margins = tableProps.getMargins(cssStyles)
        # Apply attribute "keep-with-next" ?
        kwn = cssStyles.hasClass('TableKWN') and \
              ' fo:keep-with-next="always"' or ''
        # Is the table "unbreakable" ?
        unbreak = tableProps.unbreakable and \
                  ' style:may-break-between-rows="false"' or ''
        # Do not define a specific table style if not necessary
        if not hasWidth and not margins and not kwn and not unbreak and \
           (align == 'left'):
            return 'podTable', tableWidthPx, originalTableWidthPx
        # Define a specific style for this table and return its name
        decl = '<%s:style %s:name="%s" %s:family="table" ' \
               '%s:parent-style-name="podTable"><%s:table-properties ' \
               '%s:width="%scm" %s:rel-width="%s%%" %s:table-align="%s" ' \
               '%s:align="%s"%s%s%s/></%s:style>' % \
               (s, s, self.name, s, s, s, s, formatNumber(tableWidth, sep='.'),
                s, percentage, self.tableNs, align, self.tableNs, align,
                margins, kwn, unbreak, s)
        self.env.stylesManager.dynamicStyles.add('content', decl)
        return self.name, tableWidthPx, originalTableWidthPx

    def setColumnWidth(self, width):
        '''A p_width is defined for the current cell. Store it in
           self.columnWidths'''
        # But first, ensure self.columnWidths is long enough
        widths = self.columnWidths
        while (len(widths)-1) < self.cellIndex: widths.append(None)
        # The first encountered value will be kept
        if widths[self.cellIndex] == None:
            widths[self.cellIndex] = width

    def softenWidths(self):
        '''Reduce columns widths being proportionnaly too large'''
        sizes = self.columnContentSizes
        i = len(sizes) - 1
        while i >= 0:
            # Calibrate the value in order to soften differences between small
            # and large column widths.
            value = sizes[i]
            if not value:
                i -= 1
                continue
            elif value < 5: value = 10
            elif value < 50: value = value * 2.5
            else:
                value = min(value, 1000)
            sizes[i] = int(math.pow(math.log(value), 3))
            i -= 1
        return sizes

    def computeColumnStyles(self):
        '''Once the table has been completely parsed, self.columnContentSizes
           should be correctly filled. Based on this, we can deduce the width
           of every column and create the corresponding style declarations, in
           p_self.env.stylesManager.dynamicStyles.'''
        # The objective is to compute, in "widths", relative column widths,
        # as percentages, from 0 to 1.0.
        widths = []
        i = 0
        # Compute the min and max column sizes, as percentages
        minCellWidth = min(1.0/(self.nbOfColumns*2), self.props.minColumnWidth)
        maxCellWidth = 1 - minCellWidth
        # 1st step: collect or compute widths for columns for which a width has
        # been specified.
        remainingPc = 1.0 # What global percentage will remain after this step ?
        noWidth = 0 # Count the number of columns with no specified width
        tableWidth = self.originalWidthInPx or self.widthInPx
        while i < self.nbOfColumns:
            if (i < len(self.columnWidths)) and self.columnWidths[i]:
                width = self.columnWidths[i]
                if width.unit == 'px':
                    widthAsPc = width.value / tableWidth
                elif width.unit == '%':
                    widthAsPc = width.value / 100
                else:
                    # "cm" or "pt". Ignore this for the moment.
                    widthAsPc = None
                # Ignore the computed width if wrong
                if (widthAsPc <= minCellWidth) or (widthAsPc >= maxCellWidth):
                    # A cell width of 1.0 (=100%) must be ignored: if there is a
                    # single cell, it is implicit; if there are more cells, it
                    # means that others cells will have a width of 0% and will
                    # be invisible.
                    widthAsPc = None
                widths.append(widthAsPc)
                if widthAsPc:
                    remainingPc -= widthAsPc
                else:
                    noWidth += 1
            else:
                widths.append(None)
                noWidth += 1
            i += 1
        # We must guarantee that at least 5% are available for every column
        # having no percentage yet. Else, they could be invisible. If it is not
        # the case, reset all column widths.
        if noWidth:
            required = minCellWidth * noWidth
            if remainingPc < required:
                widths = [None] * len(widths)
                remainingPc = 1.0
        # 2nd step: compute widths of columns for which no width has been
        # specified, by using self.columnContentSizes and
        # self.columnLongestWords. As a preamble, compute the total size of
        # content from all columns.
        contentTotal = 0
        i = 0
        contentSizes = self.softenWidths()
        longestWords = self.columnLongestWords
        while i < self.nbOfColumns:
            # Ignore columns for which a width has already been computed
            if widths[i] is None:
                if (i < len(contentSizes)) and contentSizes[i]:
                    contentTotal += contentSizes[i]
            i += 1
        # We will first store, in "widths", a tuple (f_width, b_minForced)
        # instead of a f_width directly. Indeed, we must ensure that every such
        # width is >= minCellWidth. All cells with a width below it will be
        # flagged with p_minForced=True. And in a second step, the surplus
        # granted to those cells will be deduced from the others.
        i = 0
        surplus = 0.0 # The total surplus granted to too narrow cells
        remainingCount = 0 # The number of cells with no surplus
        while i < self.nbOfColumns:
            if widths[i] is None:
                # Get the content size and longest word for this column
                if (i < len(contentSizes)) and contentSizes[i]:
                    contentSize = contentSizes[i]
                    longest = longestWords[i]
                else:
                    contentSize = longest = 0
                # Compute the column width
                if contentSize == 0:
                    val = 0
                else:
                    width = (float(contentSize) / contentTotal) * remainingPc
                    if width < minCellWidth:
                        surplus += minCellWidth - width
                        val = (minCellWidth, True)
                    else:
                        remainingCount += 1
                        val = (width, False)
                widths[i] = val
            i += 1
        # "Convert" stored tuples into final values
        i = 0
        while i < self.nbOfColumns:
            if isinstance(widths[i], tuple):
                if not surplus:
                    # Simply store the value without changing it
                    widths[i] = widths[i][0]
                else:
                    if widths[i][1]:
                        # Simply store the forced minimum
                        widths[i] = widths[i][0]
                    else:
                        # Reduce this value by a part of the surplus
                        widths[i] = widths[i][0] - (surplus / remainingCount)
            i += 1
        # Multiply widths (as percentages) by a predefined number, in order to
        # get a LibreOffice-compliant column width.
        i = 0
        total = 65534.0
        while i < self.nbOfColumns:
            # Compute the width of this column, relative to "total"
            widths[i] = int(widths[i] * total)
            i += 1
        # Compute style declaration corresponding to every column
        s = self.styleNs
        i = 0
        dynamic = self.env.stylesManager.dynamicStyles
        for width in widths:
            i += 1
            decl = '<%s:style %s:name="%s.%d" %s:family="table-column">' \
                   '<%s:table-column-properties %s:rel-column-width="%d*"' \
                   '/></%s:style>' % (s, s, self.name, i, s, s, s, width, s)
            dynamic.add('content', decl)

# ------------------------------------------------------------------------------
class XhtmlEnvironment(XmlEnvironment):
    itemStyles = {'ul': 'podBulletItem', 'ol': 'podNumberItem',
                  'ul_kwn': 'podBulletItemKeepWithNext',
                  'ol_kwn': 'podNumberItemKeepWithNext'}
    defaultListStyles = {'ul': 'podBulletedList', 'ol': 'podNumberedList'}
    # For list styles, this dict maps values of HTML attrbute "type" to CSS
    # property values for attribute "list-style-type".
    typeToListStyleType = {'1': 'decimal', 'a': 'lower-alpha',
      'A': 'upper-alpha', 'i': 'lower-roman', 'I': 'upper-roman',
      'disc': 'disc', 'circle': 'circle', 'square': 'square'}
    # Mapping between HTML list styles and ODT list styles
    listClasses = {'ol': NumberedProperties, 'ul': BulletedProperties}
    # "list-style-type" values supported by OpenDocument
    listFormats = {
      # Number formats
      'lower-alpha': ('a',), 'upper-alpha': ('A',), 'lower-latin': ('a',),
      'upper-latin': ('A',), 'lower-roman': ('i',), 'upper-roman': ('I',),
      'decimal': ('1',),
      # Bullet formats
      'disc': (u'•'), 'circle': (u'◦'), 'square': (u'▪'), 'none': ''}

    def __init__(self, renderer):
        XmlEnvironment.__init__(self)
        self.renderer = renderer
        self.stylesManager = renderer.stylesManager
        self.ns = renderer.currentParser.env.namespaces
        self.res = u''
        self.currentContent = u''
        self.currentElements = [] # Stack of currently walked elements
        self.currentLists = [] # Stack of currently walked lists (ul or ol)
        self.currentTables = [] # Stack of currently walked tables
        self.lastElem = None # Last walked element before the current one
        self.textNs = self.ns[OdfEnvironment.NS_TEXT]
        self.linkNs = self.ns[OdfEnvironment.NS_XLINK]
        self.tableNs = self.ns[OdfEnvironment.NS_TABLE]
        self.styleNs = self.ns[OdfEnvironment.NS_STYLE]
        # The following attr will be True when parsing parts of the XHTML that
        # must be ignored.
        self.ignore = False
        # Maintain a dict of collected [Bulleted|Numbered]Properties instances,
        # to avoid generating the corresponding list styles several times.
        self.listProperties = {}
        # The fusion of currently encountered styles corresponding to XHTML
        # "style-only" tags (STYLE_ONLY_TAGS).
        self.mergedInnerStyles = CssStyles()
        # How much STYLE_ONLY_TAGS are currently walked (whose fusion is in
        # self.mergedInnerStyles) ?
        self.mergedCount = 0
        # The width (in cm, as a float) of pages (margins excluded) within the
        # current document.
        self.pageWidth = renderer.stylesManager.pageLayout.getWidth()
        # It may be required to enable the POD post-processor (=PPP). But did we
        # already do it ?
        self.pppEnabled = False

    def enablePpp(self):
        '''Enable the POD post-processor (if not already done yet)'''
        if self.pppEnabled: return
        self.renderer.enablePpp()
        self.pppEnabled = True

    def getUniqueStyleName(self, type):
        '''Gets a unique name for an element of some p_type (a table, a
           list).'''
        return getUuid(removeDots=True, prefix=type.capitalize())

    def getCurrentElement(self, isList=False):
        '''Gets the element that is on the top of self.currentElements or
           self.currentLists.'''
        res = None
        if isList:
            elements = self.currentLists # Stack of list elements only
        else:
            elements = self.currentElements # Stack of all elements (including
            # elements also pushed on other stacks, like lists and tables).
        if elements:
            res = elements[-1]
            if res.removeTag:
                # This tag will not be dumped: the real current one is the
                # parent.
                res = res.parent
        return res

    def anElementIsMissing(self, previous, current):
        return previous and (previous.elem in OUTER_TAGS) and \
               ((not current) or (current.elem in INNER_TAGS))

    def dumpCurrentContent(self, place, elem):
        '''Dumps content that was temporarily stored in self.currentContent
           into the result.'''
        current = self.getCurrentElement()
        # Remove the trailing whitespace if needed
        if place == 'start':
            if self.currentContent.endswith(' ') and \
               ((elem not in INNER_TAGS_NO_BR) or (elem == 'img')):
                self.currentContent = self.currentContent[:-1]
        # Remove the leading whitespace if needed
        if self.currentContent.startswith(' '):
            last = self.lastElem
            if (place == 'end') and (elem in INNER_TAGS):
                parent = current.parent
                if not parent or \
                   ((parent.elem not in INNER_TAGS) and \
                     not parent.contentDumped) or \
                   (last and (last.elem == 'br')):
                    self.currentContent = self.currentContent[1:]
            else:
                if not last or (last.elem not in INNER_TAGS_NO_BR):
                    self.currentContent = self.currentContent[1:]
        if self.currentContent:
            # Manage missing elements
            if self.anElementIsMissing(current, None):
                current.addInnerParagraph(self)
            # Dump the current content
            self.dumpString(Escape.xml(self.currentContent))
            current.show(self, content='...', prefix='')
            for e in self.currentElements:
                e.contentDumped = True
        # Reinitialise the current content
        if self.currentContent:
            # If we are within a table cell, update the total size of cell
            # content and the longest found word.
            if self.currentTables and self.currentTables[-1].inCell:
                contentSize = len(self.currentContent)
                longest = 0 # Longest word's size
                # "longest" is currently not used. We do not compute it to save
                # processing.
                # for word in self.currentContent.split():
                #    longest = max(longest, len(word))
                for table in self.currentTables:
                    table.cellContentSize += contentSize
                    table.cellLongestWord = max(table.cellLongestWord, longest)
            self.currentContent = u''

    def getOdfAttributes(self, xhtmlElem):
        '''Gets the ODF attributes to dump for p_xhtmlElem'''
        # Complete attributes if inherited from a parent tag
        if xhtmlElem.elem in ('td', 'th'):
            table = self.currentTables[-1]
            # Manage cells' borders
            border = table.props.border
            if border is None:
                border = table.border and '0.05pt solid #000000' or '0'
            xhtmlElem.cssStyles.add('border', border)
            # Manage cells' spacing
            spacing = table.borderSpacing
            if spacing:
                # A minimum cell padding may be applicable
                value = table.props.getCellPadding(spacing)
                xhtmlElem.cssStyles.add('border-spacing', '%.2fcm' % value)
        else:
            parent = xhtmlElem.parent
            if parent and (parent.elem in ('td', 'th')) and \
               (xhtmlElem.elem in PARA_TAGS):
                # The enclosed "p" must get the td's inner styles
                if parent.innerStyle:
                    xhtmlElem.cssStyles.addClass(parent.innerStyle,
                                                 append=False)
                if parent.innerCssStyles:
                    xhtmlElem.cssStyles.merge(parent.innerCssStyles)
        # Get the base ODF style for this p_xhtmlElem
        style = self.findStyle(xhtmlElem)
        if style: return style.getOdfAttributes(xhtmlElem.attrs)
        return ''

    def addListProperties(self, elem, listProps):
        '''Ensures the ListProperties instance p_listProps is among
           self.listProperties.'''
        for name, value in self.listProperties.iteritems():
            if value == listProps:
                return name
        # If we are here, p_listProps wat not found. Add it.
        name = self.getUniqueStyleName('list')
        self.listProperties[name] = listProps
        return name

    def addListPropertiesByName(self, elem, name):
        '''A specific list style must be used, based on HTML attribute "type" or
           CSS property "list-style-type". Add the corresponding
           BulletedProperties or NumberedProperties instance in
           self.listProperties if not present yet.'''
        if name in self.listProperties: return
        # Determine the ListProperties class to use
        klass = self.listClasses[elem]
        # Determine the format of bullets/numbers
        formats = (name in self.listFormats) and self.listFormats[name] or \
                  klass.defaultFormats
        # Create the Properties instance
        styleName = 'L-%s' % name
        self.listProperties[styleName] = klass(formats=formats)
        return styleName

    def getListStyle(self, xhtmlElem, attrs):
        '''Gets the list style to apply to p_xhtmlElem (a "ol" or "ul"). If no
           specific style information is found in p_attrs, a default style is
           applied, from XhtmlEnvironment.defaultListStyles. Else, a specific
           list style is created and added to dynamic styles.'''
        # Check if a specific style must be created
        res = None
        # Check first in the styles mappings
        listProps = self.findStyle(xhtmlElem)
        elem = xhtmlElem.elem
        if listProps:
            # I have a ListProperties instance from a styles mapping. Get a name
            # for the corresponding style.
            xhtmlElem.paraStyle = listProps.paraStyle
            return self.addListProperties(elem, listProps)
        # Check CSS attribute "list-style-type"
        styles = xhtmlElem.cssStyles
        if hasattr(styles, 'liststyletype'):
            typeValue = styles.liststyletype.value
            if typeValue not in ('initial', 'inherit'):
                res = typeValue
        # Check HTML attribute "type"
        if not res and attrs.has_key('type'):
            res = self.typeToListStyleType.get(attrs['type'])
        if res:
            # A specific style has been found. Ensure it will be added among
            # dynamic styles.
            res = self.addListPropertiesByName(elem, res)
        else:
            # Apply a default style, added by default among dynamic styles
            res = self.defaultListStyles[elem]
        return res

    def getTags(self, elems, start=True,
                ignoreToRemove=False, ignoreWaiting=False):
        '''This method returns a series of start or end tags (depending on
           p_start) that correspond to HtmlElement instances in p_elems.'''
        res = ''
        for elem in elems:
            # Ignore tags flagged "to remove" when relevant
            if ignoreToRemove and elem.removeTag: continue
            # Ignore not-yet-dumped elements when relevant
            if ignoreWaiting and (elem.dumpStatus == 'waiting'): continue
            # Get the tag
            tag = elem.dump(start, self)
            if start: res += tag
            else: res = tag + res
        return res

    def closeConflictualElements(self, conflictElems):
        '''This method dumps end tags for p_conflictElems, excepted if those
           tags would be empty. In this latter case, tags are purely removed
           from the result.'''
        startTags = self.getTags(conflictElems, start=True,
                                 ignoreToRemove=True, ignoreWaiting=True)
        if startTags and self.res.endswith(startTags):
            # In this case I would dump an empty (series of) tag(s). Instead, I
            # will remove those tags.
            self.res = self.res[:-len(startTags)]
        else:
            tags = self.getTags(conflictElems, start=False,
                                ignoreToRemove=True, ignoreWaiting=True)
            self.dumpString(tags)
        return conflictElems

    def dumpString(self, s):
        '''Dumps arbitrary content p_s.
           If the table stack is not empty, we must dump p_s into the buffer
           corresponding to the last parsed table. Else, we must dump p_s
           into the global buffer (self.res).'''
        if self.currentTables:
            currentTable = self.currentTables[-1]
            if (not currentTable.res) or currentTable.firstRowParsed:
                currentTable.res += s
            else:
                currentTable.tempRes += s
        else:
            self.res += s

    def getTagsToReopen(self, conflictElems):
        '''Normally, tags to reopen are equal to p_conflictElems. But we have a
           special case. Indeed, if a conflict elem has itself tagsToClose,
           the last tag to close may not be needed anymore on the tag to
           reopen, so we remove it.'''
        conflictElems[-1].tagsToClose = None
        return conflictElems

    def onElementStart(self, elem, attrs):
        '''Returns an HtmlElement instance representing the currently walked
           p_elem.'''
        self.dumpCurrentContent('start', elem)
        parent = self.getCurrentElement()
        current = HtmlElement(elem, attrs, parent=parent)
        # Insert a line-break when relevant
        if parent and parent.dumpLineBreakOnInner:
            if elem in INNER_TAGS:
                self.dumpString('<text:line-break/>')
                current.show(self, content='br', prefix='')
            parent.dumpLineBreakOnInner = False
        # Manage conflictual elements
        conflictElems = current.getConflictualElements(self)
        if conflictElems:
            # We must close the conflictual elements, and once the currentElem
            # will be dumped, we will re-open the conflictual elements.
            toReopen = self.closeConflictualElements(conflictElems)
            current.tagsToReopen = self.getTagsToReopen(toReopen)
        # Manage missing elements
        if self.anElementIsMissing(parent, current):
            parent.addInnerParagraph(self)
        # Add the current element to the stack of walked elements
        self.currentElements.append(current)
        if elem in XHTML_LISTS:
            # Update stack of current lists
            self.currentLists.append(current)
        elif elem == 'table':
            # Update stack of current tables
            self.currentTables.append(HtmlTable(self, current, attrs))
        elif elem in TABLE_COL_TAGS:
            # Determine colspan
            colspan = 1
            if attrs.has_key('colspan'): colspan = int(attrs['colspan'])
            table = self.currentTables[-1]
            table.inCell = colspan
            table.cellIndex += colspan
            # If we are in the first row of a table, update columns count
            if not table.firstRowParsed:
                table.nbOfColumns += colspan
            styles = current.cssStyles
            if hasattr(styles, 'width') and (colspan == 1):
                table.setColumnWidth(styles.width)
            # Determine the styles to apply to inner-cell paragraphs
            if elem == 'td':
                current.innerStyle = table.props.cellContentStyle
            elif elem == 'th':
                current.innerStyle = table.props.headerContentStyle
        return current

    def onElementEnd(self, elem):
        res = None
        self.dumpCurrentContent('end', elem)
        current = self.currentElements.pop()
        if elem in XHTML_LISTS:
            self.currentLists.pop()
        elif elem == 'table':
            table = self.currentTables.pop()
            if table.nbOfColumns:
                # Computes the column styles required by the table
                table.computeColumnStyles()
            # Dumps the content of the last parsed table into the parent buffer
            self.dumpString(table.res)
        elif elem in TABLE_ROW_TAGS:
            table = self.currentTables[-1]
            table.cellIndex = -1
            if not table.firstRowParsed and table.nbOfColumns:
                # If no cell has been parsed yet, we have an empty row that must
                # be ignored.
                table.firstRowParsed = True
                # First row is parsed. The number of columns in the table is
                # known: columns declarations can be dumped.
                for i in range(1, table.nbOfColumns + 1):
                    table.res+= '<%s:table-column %s:style-name="%s.%d"/>' % \
                                (self.tableNs, self.tableNs, table.name, i)
                table.res += table.tempRes
                table.tempRes = u''
        elif elem in TABLE_COL_TAGS:
            table = self.currentTables[-1]
            # If we are walking a td or th, update "columnContentSizes" and
            # "columnLongestWords" for the currently parsed table, excepted if
            # the cell spans several columns.
            if elem != 'col':
                # Divide the content size if cell colspan > 1
                cellContentSize = table.cellContentSize / table.inCell
                cellLongestWord = table.cellLongestWord / table.inCell
                i = table.cellIndex
                sizes = table.columnContentSizes
                wordSizes = table.columnLongestWords
                # Insert None values if the lists are too small
                while (len(sizes)-1) < i:
                    sizes.append(None)
                    wordSizes.append(None)
                longest = max(sizes[i], cellContentSize, 5)
                wordLongest = max(wordSizes[i], cellLongestWord, 5)
                sizes[i] = longest
                wordSizes[i] = min(wordLongest, 25)
            table.inCell = table.cellContentSize = table.cellLongestWord = 0
        if current.tagsToClose:
            self.closeConflictualElements(current.tagsToClose)
        if current.tagsToReopen:
            res = current.tagsToReopen
        if current.removeTag:
            current.parent.dumpLineBreakOnInner = True
        self.lastElem = current
        return current, res

    def findStyle(self, elem):
        converter = self.parser.caller
        localStylesMapping = converter.localStylesMapping
        return converter.stylesManager.findStyle(elem, localStylesMapping)

    def updateMergedStyles(self, action, xhtmlElem):
        '''Updates the current set of applicable inner styles after inner tag
           p_xhtmlElem has been encountered.'''
        styles = self.mergedInnerStyles
        if action == 'add':
            # Add the CSS styles related to the encountered p_xhtmlElem to
            # self.mergedInnerStyles.
            originalStyles = xhtmlElem.cssStyles
            self.mergedInnerStyles.merge(originalStyles)
            xhtmlElem.cssStyles = self.mergedInnerStyles
            xhtmlElem.originalStyles = originalStyles
            self.mergedCount += 1
        elif action == 'delete':
            # Remove the CSS styles related to p_xhtmlElem from
            # self.mergedInnerStyles.
            originalStyles = xhtmlElem.originalStyles
            self.mergedInnerStyles.unmix(originalStyles)
            xhtmlElem.cssStyles = originalStyles
            xhtmlElem.originalStyles = None
            self.mergedCount -= 1

# ------------------------------------------------------------------------------
class XhtmlParser(XmlParser):
    def lowerizeInput(self, elem, attrs=None):
        '''Because (X)HTML is case insensitive, we may receive input p_elem and
           p_attrs in lower-, upper- or mixed-case. So here we produce lowercase
           versions that will be used throughout our parser.'''
        resElem = elem.lower()
        resAttrs = attrs
        if attrs:
            resAttrs = {}
            for attrName in attrs.keys():
                resAttrs[attrName.lower()] = attrs[attrName]
        if attrs == None:
            return resElem
        else:
            return resElem, resAttrs

    def startElement(self, elem, attrs):
        elem, attrs = self.lowerizeInput(elem, attrs)
        e = self.env
        dump = e.dumpString
        current = e.onElementStart(elem, attrs)
        parent = current.parent
        current.show(e)
        if current.removeTag: # Do not dump this tag
            if parent.contentDumped:
                # Dump a line break instead, only if content was already dumped
                # into the parent.
                dump('<text:line-break/>')
                current.show(self, content='br', prefix='')
            return
        odfTag = current.getOdfTag()
        if elem in SIMPLE_TAGS:
            dump('<' + odfTag)
            dump(e.getOdfAttributes(current))
            dump('>')
        elif elem in STYLE_ONLY_TAGS:
            # Dump an end tag if we are already dumping styled content
            if e.mergedCount and (parent.dumpStatus == 'dumped'):
                dump('</%s>' % odfTag)
            e.updateMergedStyles('add', current)
            # We will dump this tag only if subsequent content is found
            current.dumpStatus = 'waiting'
        elif elem == 'a':
            dump('<%s %s:type="simple"' % (odfTag, e.linkNs))
            if attrs.has_key('href'):
                dump(' %s:href="%s"' % (e.linkNs, Escape.xml(attrs['href'])))
            dump('>')
        elif elem in XHTML_LISTS:
            prologue = ''
            if parent and (parent.elem in XHTML_LISTS):
                # It is a list into another list. In this case the inner list
                # must be surrounded by a list-item element.
                prologue = '<text:list-item>'
            numbering = ''
            if elem == 'ol':
                # By default, LO seems to continue numbering or previous lists
                numbering = ' text:continue-numbering="false"'
            current.odStyle = e.getListStyle(current, attrs)
            dump('%s<%s text:style-name="%s"%s>' % (
                 prologue, odfTag, current.odStyle, numbering))
        elif elem == 'li':
            # Must numbering restart at this "li" ?
            current = e.currentLists[-1]
            current.subNumber += 1
            numbered = current.elem == 'ol'
            start = None
            if numbered:
                attrs = current.attrs
                start = None
                if attrs and ('start' in attrs):
                    start = attrs['start']
                    del attrs['start']
                elif current.subNumber == 1:
                    start = 1
            if start is None:
                restart = ''
            else:
                restart = ' text:start-value="%d"' % start
            dump('<%s%s>' % (odfTag, restart))
        elif elem == 'thead':
            dump('<%s>' % odfTag)
        elif elem == 'tr':
            attr = e.currentTables[-1].props.unbreakableRows and \
                   ' table:style-name="podUnbreakableRow"' or ''
            dump('<%s%s>' % (odfTag, attr))
        elif elem == 'table':
            # Here we must call "dumpString" only once
            table = e.currentTables[-1]
            dump('<%s %s:name="%s" %s:style-name="%s">' % \
                 (odfTag, e.tableNs, table.name, e.tableNs, table.style))
        elif elem == 'img':
            conv = self.caller
            src = attrs.get('src', '').strip()
            if src: 
                imgCode = e.renderer.importDocument(at=attrs['src'],
                  format='image', wrapInPara=False, style=current.cssStyles,
                  maxWidth=conv.imagesMaxWidth, keepRatio=conv.keepImagesRatio)
                dump(imgCode)
        elif elem == 'footnote':
            # An inexistent tag in HTML, nervertheless usable to produce ODF
            # footnotes.
            nb = attrs['nb']
            dump('<text:note text:id="ftn%s" text:note-class="footnote">' \
                 '<text:note-citation>%s</text:note-citation><text:note-body>'%\
                 (nb, nb))
        elif elem == 'tab':
            # Another inexistent tag in HTML, useful for managing tabs
            dump('<text:tab/>')
        elif elem in IGNORABLE_TAGS:
            e.ignore = True

    def endElement(self, elem):
        elem = self.lowerizeInput(elem)
        e = self.env
        dump = e.dumpString
        current, elemsToReopen = e.onElementEnd(elem)
        # Determine the tag to dump
        startTag, endTag = current.getOdfTags()
        if current.isConflictual:
            # Compute the start tag, with potential styles applied
            startTag = e.getTags((current,), start=True)
        if current.isConflictual and e.res.endswith(startTag):
            # We will not dump it, it would constitute a silly empty tag
            e.res = e.res[:-len(startTag)]
            if current.elem in STYLE_ONLY_TAGS:
                e.mergedCount -= 1
        else:
            # Dump the end tag, but dump some additional stuff if required
            if elem == 'footnote':
                dump('</text:note-body></text:note>')
            elif elem in XHTML_LISTS:
                if current.parent and (current.parent.elem in XHTML_LISTS):
                    # We were in an inner list. So we must close the list-item
                    # tag that surrounds it.
                    endTag = '%s</text:list-item>' % endTag
            if endTag and not current.removeTag and \
               (current.dumpStatus != 'waiting'):
                dump(endTag)
                current.show(e, prefix='>')
            # Manage the end of a styled inner tag
            if elem in STYLE_ONLY_TAGS:
                # Unmix styles corresponding to this end tag
                e.updateMergedStyles('delete', current)
                parent = current.parent
                if e.mergedCount and (parent.elem in STYLE_ONLY_TAGS):
                    # Possibly reopen a tag. The parent is already loaded with
                    # merged styles.
                    parent.dumpStatus = 'waiting'
            elif (elem not in INNER_TAGS) and not current.inInnerTag():
                # We are not in an "inner" zone anymore
                e.mergedCount = 0
        if elem in IGNORABLE_TAGS:
            e.ignore = False
        if elemsToReopen:
            dump(e.getTags(elemsToReopen, start=True, ignoreWaiting=True))

    def characters(self, content):
        e = XmlParser.characters(self, content)
        if e.ignore: return
        # Dump a tag waiting for content
        current = e.getCurrentElement()
        if current.dumpStatus == 'waiting':
            dump = e.dumpString
            dump('<' + current.getOdfTag())
            dump(e.getOdfAttributes(current))
            dump('>')
            current.dumpStatus = 'dumped'
        e.currentContent += WhitespaceCruncher.crunch(content, e.currentContent)

    def endDocument(self):
        '''Dump all collected list styles'''
        styles = self.caller.stylesManager.dynamicStyles.styles_base
        for name, props in self.env.listProperties.iteritems():
            styles.append(props.dumpStyle(name))

# ------------------------------------------------------------------------------
class XhtmlPreprocessor:
    '''The Xhtml2OdtConverter (see below) requires valid XHTML as input, because
       it uses a SAX parser to do its job. This preprocessor ensures it is the
       case, by applying some modifications to this input.'''
    # Regular expression representing a HTML void tag
    voidTags = ('area', 'base', 'br', 'col', 'hr', 'img', 'input', 'link',
                'meta', 'param', 'command', 'keygen', 'source')
    voidTag = re.compile('<(%s)([^>]*?)(/)?\s*>' % '|'.join(voidTags), re.S)

    @staticmethod
    def patchTag(match):
        '''Add the trailing slash if absent'''
        # If char "/" is already here, do not change anything
        if match.group(3): return match.group(0)
        return '<%s%s/>' % (match.group(1), match.group(2))

    @staticmethod
    def html2xhtml(s):
        '''Converts valid HTML p_s to valid XHTML'''
        return re.sub(XhtmlPreprocessor.voidTag, XhtmlPreprocessor.patchTag, s)

    @staticmethod
    def preprocess(s, html=False):
        '''Converts string p_s to valid XHTML and r_eturns it. If p_html is
           True, p_s is supposed to be valid HTML: void tags must be converted
           to XHTML's self-closing start tags.'''
        if s == None: s = ''
        # Manage encoding
        if isinstance(s, unicode): s = s.encode('utf-8')
        # Surround p_s with a tag in order to get a XML-compliant file (we need
        # a root tag). We also remove special blank chars that produce SAX
        # parsing errors.
        s = '<p>%s</p>' % s.replace('\f', '').replace('\v', '')
        # Convert, when required, conversion of HTML's void tags to XHTML's
        # self-closing start tags.
        if html:
            s = XhtmlPreprocessor.html2xhtml(s)
        return s

# ------------------------------------------------------------------------------
class Xhtml2OdtConverter:
    '''Converts a chunk of XHTML into a chunk of ODT'''
    verbose = False

    def __init__(self, s, encoding, stylesManager, localStylesMapping,
                 keepWithNext, keepImagesRatio, imagesMaxWidth, renderer,
                 html, unwrap):
        self.renderer = renderer
        self.xhtmlString = XhtmlPreprocessor.preprocess(s, html=html)
        self.encoding = encoding # Todo: manage encoding that is not utf-8
        self.stylesManager = stylesManager
        self.localStylesMapping = localStylesMapping
        self.odtChunk = None
        self.xhtmlParser = XhtmlParser(XhtmlEnvironment(renderer), self)
        if keepWithNext:
            if (keepWithNext is True) or (keepWithNext > 1):
                # Apply "keep-with-next" functionality by using a sub-SAX parser
                from appy.pod.xhtml import parser as xparser
                parser = xparser.XhtmlParser(xparser.XhtmlEnvironment(), self,
                                             keepWithNext=keepWithNext)
                self.xhtmlString= parser.parse(self.xhtmlString).encode('utf-8')
            else:
                self.xhtmlString = self.applyKeepWithNext()
        self.keepImagesRatio = keepImagesRatio
        self.imagesMaxWidth = imagesMaxWidth
        self.unwrap = unwrap
        # In verbose mode, we dump a trace of the xhtml2odt algorithm
        self.xhtmlParser.verbose = self.xhtmlParser.env.verbose = self.verbose
        # Some components within the styles manager may require the XHTML parser
        # for generating some styles.
        self.stylesManager.setXhtmlParser(self.xhtmlParser)

    def run(self):
        '''Parses the input XHTML string and returns the resulting ODF chunk'''
        self.xhtmlParser.parse(self.xhtmlString)
        r = self.xhtmlParser.env.res
        if self.unwrap and r.startswith('<text:p>'):
            r = r[8:-9]
        return r

    def applyKeepWithNext(self):
        '''This method is called prior to parsing self.xhtmlString in order to
           add specific CSS classes to some XHTML tags, implementing the
           "keep-with-next" functionality. If the last tag is:
           * a paragraph (tag "p"), class "ParaKWN" will be set;
           * a bullet (tag "li"), class "podItemKeepWithNext" will be set.

           Note that this latter class will then be converted by the XHTML
           parser into "real" style "podBulletItemKeepWithNext" or
           "podNumberItemKeepWithNext", if the "li" is, respectively, in a "ul"
           or "ol" tag.
        '''
        res = self.xhtmlString
        lastParaIndex = res.rfind('<p')
        lastItemIndex = res.rfind('<li')
        if (lastParaIndex != -1) or (lastItemIndex != -1):
            # Is the last one a paragraph or an item ?
            if lastParaIndex > lastItemIndex:
                # A paragraph
                styleName = 'ParaKWN'
                elemLenght = 2
            else:
                # An item
                styleName = 'podItemKeepWithNext'
                elemLenght = 3
            maxIndex = max(lastParaIndex, lastItemIndex)
            # Does this element already have a "class" attribute ?
            if res.find('class="', maxIndex) == -1:
                # No: I add the style
                res = res[:maxIndex+elemLenght] + (' class="%s" ' % styleName) \
                      + res[maxIndex+elemLenght:]
        return res
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy import Object
from appy.px import Px

# ------------------------------------------------------------------------------
class Phase:
    '''A group of pages'''

    pxView = Px('''
     <table class="phase"
            var="singlePage=len(phase.pages) == 1;
                 label='%s_phase_%s' % (zobj.meta_type, phase.name)">
      <tr valign="top">
       <!-- The page(s) within the phase -->
       <td for="aPage in phase.pages"
           var2="aPageInfo=phase.pagesInfo[aPage]"
           class=":(aPage == page) and 'currentPage' or ''">
        <!-- First line: page name and icons -->
        <span if="not (singlePhase and singlePage)">
         <x var="label=aPageInfo.page.getLabel(zobj)">
          <a if="aPageInfo.showOnView"
             href=":zobj.getUrl(page=aPage, inPopup=inPopup)">::label</a>
          <x if="not aPageInfo.showOnView">:label</x>
         </x>
         <x var="locked=zobj.isLocked(user, aPage);
                 editable=mayEdit and aPageInfo.showOnEdit and \
                          aPageInfo.showEdit">
          <a if="editable and not locked"
             href=":zobj.getUrl(mode='edit', page=aPage, inPopup=inPopup)">
           <img src=":url('edit')" title=":_('object_edit')"/></a>
          <a if="editable and locked">
           <img style="cursor: help"
                var="lockDate=ztool.formatDate(locked[1]);
                     lockMap={'user':ztool.getUserName(locked[0]), \
                              'date':lockDate};
                     lockMsg=_('page_locked', mapping=lockMap)"
                src=":url('locked')" title=":lockMsg"/></a>
          <a if="editable and locked and user.has_role('Manager')">
           <img class="clickable" title=":_('page_unlock')" src=":url('unlock')"
                onclick=":'onUnlockPage(%s,%s)' % (q(zobj.id), q(aPage))"/></a>
         </x>
        </span>
        <!-- Next lines: links -->
        <x var="links=aPageInfo.links" if="links">
         <div for="link in links" class="refLink">
          <a href=":link.url">:link.title</a></div>
        </x>
       </td>
      </tr>
     </table>''')

    # "Static" PX displaying all phases of a given object
    pxAllPhases = Px('''
     <x var="singlePhase=len(phases)==1;
             page=req.get('page', '');
             uid=zobj.id;
             mayEdit=zobj.mayEdit()">
      <x if="singlePhase" var2="phase=phases[0]">:phase.pxView</x>
      <!-- Display several phases in tabs -->
      <x if="not singlePhase">
       <table cellpadding="0" cellspacing="0">
        <!-- First row: the tabs -->
        <tr><td class="tabSep">
         <table class="tabs" id=":'tabs_%s' % uid">
          <tr valign="middle">
           <x for="phase in phases"
              var2="suffix='%s_%s' % (uid, phase.name);
                    tabId='tab_%s' % suffix">
            <td id=":tabId" class="tab">
             <a onclick=":'showTab(%s)' % q(suffix)"
                class="clickable">:_('%s_phase_%s' % (zobj.meta_type, \
                                                      phase.name))</a>
            </td>
           </x>
          </tr>
         </table>
        </td></tr>
        <!-- Other rows: the fields -->
        <tr for="phase in phases" id=":'tabcontent_%s_%s' % (uid, phase.name)"
          style=":(loop.phase.nb==0) and 'display:table-row' or 'display:none'">
         <td>:phase.pxView</td>
        </tr>
       </table>
       <script type="text/javascript">:'initTab(%s,%s)' % \
        (q('tab_%s' % uid), q('%s_%s' % (uid, phases[0].name)))
       </script>
      </x>
     </x>''')

    def __init__(self, name, obj):
        self.name = name
        self.obj = obj
        # The list of names of pages in this phase
        self.pages = []
        # The list of hidden pages in this phase
        self.hiddenPages = []
        # The dict below stores info about every page listed in self.pages.
        self.pagesInfo = {}
        self.totalNbOfPhases = None
        # The following attributes allows to browse, from a given page, to the
        # last page of the previous phase and to the first page of the following
        # phase if allowed by phase state.
        self.previousPhase = None
        self.nextPhase = None

    def addPageLinks(self, field, obj):
        '''If p_field is a navigable Ref, we must add, within self.pagesInfo,
           objects linked to p_obj through this Ref as links.'''
        if field.page.name in self.hiddenPages: return
        infos = []
        for ztied in field.getValue(obj, appy=False):
            infos.append(Object(title=ztied.title, url=ztied.absolute_url()))
        self.pagesInfo[field.page.name].links = infos

    def addPage(self, field, obj, layoutType):
        '''Adds page-related information in the phase.'''
        # If the page is already there, we have nothing more to do.
        if (field.page.name in self.pages) or \
           (field.page.name in self.hiddenPages): return
        # Add the page only if it must be shown.
        showOnView = field.page.isShowable(obj, 'view')
        showOnEdit = field.page.isShowable(obj, 'edit')
        if showOnView or showOnEdit:
            # The page must be added
            self.pages.append(field.page.name)
            # Create the dict about page information and add it in self.pageInfo
            pageInfo = Object(page=field.page, showOnView=showOnView,
                              showOnEdit=showOnEdit, links=None)
            pageInfo.update(field.page.getInfo(obj, layoutType))
            self.pagesInfo[field.page.name] = pageInfo
        else:
            self.hiddenPages.append(field.page.name)

    def computeNextPrevious(self, allPhases):
        '''This method also fills fields "previousPhase" and "nextPhase"
           if relevant, based on list of p_allPhases.'''
        # Identify previous and next phases
        for phase in allPhases:
            if phase.name == self.name:
                i = allPhases.index(phase)
                if i > 0:
                    self.previousPhase = allPhases[i-1]
                if i < (len(allPhases)-1):
                    self.nextPhase = allPhases[i+1]

    def getPreviousPage(self, page):
        '''Returns the page that precedes p_page in this phase'''
        try:
            pageIndex = self.pages.index(page)
        except ValueError:
            # The current page is probably not visible anymore. Return the
            # first available page in current phase.
            if self.pages:
                res = self.pages[0]
                return res, self.pagesInfo[res]
            return None, None
        if pageIndex > 0:
            # We stay on the same phase, previous page
            res = self.pages[pageIndex-1]
            return res, self.pagesInfo[res]
        else:
            if self.previousPhase:
                # We go to the last page of previous phase
                previousPhase = self.previousPhase
                res = previousPhase.pages[-1]
                return res, previousPhase.pagesInfo[res]
            else:
                return None, None

    def getNextPage(self, page):
        '''Returns the page that follows p_page in this phase.'''
        try:
            pageIndex = self.pages.index(page)
        except ValueError:
            # The current page is probably not visible anymore. Return the
            # first available page in current phase.
            if self.pages:
                res = self.pages[0]
                return res, self.pagesInfo[res]
            return None, None
        if pageIndex < (len(self.pages)-1):
            # We stay on the same phase, next page
            res = self.pages[pageIndex+1]
            return res, self.pagesInfo[res]
        else:
            if self.nextPhase:
                # We go to the first page of next phase
                nextPhase = self.nextPhase
                res = nextPhase.pages[0]
                return res, nextPhase.pagesInfo[res]
            else:
                return None, None

    def getPageInfo(self, page, layoutType):
        '''Return the page info corresponding to the given p_page. If this page
           cannot be shown on p_layoutType, this method returns page info about
           the first showable page on p_layoutType, or None if no page is
           showable at all.'''
        res = self.pagesInfo[page]
        showAttribute = 'showOn%s' % layoutType.capitalize()
        if getattr(res, showAttribute): return res
        # Find the first showable page in this phase on p_layoutType.
        for pageName in self.pages:
            if pageName == page: continue
            pageInfo = self.pagesInfo[pageName]
            if getattr(pageInfo, showAttribute): return pageInfo
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# This module implements "integration" testing. It works by:
# - defining an executable test plan as an ODT file containing tests as tables
#   conforming to conventions as described in appy/utils/tables.py ;
# - executing this plan, or only a sub-part of its tests ;
# - producing a report about success or failure of every test ran.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path
import os, os.path, sys, time

from appy.bin import Program
from appy.xml.diff import Diff
from appy.utils import Traceback
from appy.utils import path as putils
from appy.utils.tables import TablesParser

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TesterError(Exception): pass
class InternalError(Exception): pass

# TesterError-related constants
WRONG_TEST_PLAN = 'The test plan you specified does not correspond to an ' \
  'existing ODT file.'
_FLAVOUR = 'A flavour represents a test configuration.'
FLAVOURS_NOT_LIST = 'The flavours specified must be a list or tuple of ' \
  'string. ' + _FLAVOUR
FLAVOUR_NOT_STRING = 'Each specified flavour must be a string. ' + _FLAVOUR
WRONG_TEST_FACTORY = 'You must give a test factory that inherits from the ' \
  'abstract "appy.shared.test.TestFactory" class.'
CREATE_TEST_NOT_OVERRIDDEN = 'The appy.shared.test.TestFactory.createTest ' \
  'method must be overridden in your concrete TestFactory.'
MAIN_TABLE_NOT_FOUND = 'No table "TestSuites" found in test plan "%s".'
MAIN_TABLE_MALFORMED = 'The "TestSuites" table must have at least two ' \
  'columns, named "Name" and "Description".'
TEST_SUITE_NOT_FOUND = 'Table "%s.descriptions" and/or "%s.data" were not ' \
  'found.'
TEST_SUITE_MALFORMED = 'Tables "%s.descriptions" and "%s.data" do not have ' \
  'the same length. For each test in "%s.data", You should have one line in ' \
  '"%s.descriptions" describing the test.'
FILE_NOT_FOUND = 'File to compare "%s" was not found.'
WRONG_FLAVOUR = 'Wrong flavour "%s". Flavour must be one of %s.'
TEST_AND_SUITE = 'Specify a single test or test suite but not both.'

# InternalError-related constants
TEST_REPORT_SINGLETON_ERROR = 'You can only use the TestReport constructor ' \
  'once. After that you can access the single TestReport instance via the ' \
  'TestReport.instance static member.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TestReport:
    '''Collects data about test results for producing a report'''

    # The single TestReport instance to be ever created
    instance = None

    def __init__(self, reportPath, verbose):
        if TestReport.instance is None:
            self.report = open(str(reportPath), 'w')
            self.verbose = verbose
            TestReport.instance = self
        else:
            raise InternalError(TEST_REPORT_SINGLETON_ERROR)

    def say(self, msg, force=False):
        if self.verbose or force:
            print(msg)
        self.report.write(msg)
        self.report.write('\n')

    def close(self): self.report.close()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Test:
    '''Abstract test class'''
    def __init__(self, testData, testDescription, testFolder,
                 config, flavour, rendererParams):
        self.data = testData
        self.description = testDescription
        self.testFolder = testFolder
        self.tempFolder = None
        self.report = TestReport.instance
        self.errorDump = None
        self.config = config
        self.flavour = flavour
        self.rendererParams = rendererParams

    def compareFiles(self, expected, actual, areXml=False, xmlTagsToIgnore=(),
                     xmlAttrsToIgnore=()):
        '''Compares 2 files. r_ is True if files are different. The differences
        are written in the test report.'''
        for f in expected, actual:
            assert os.path.exists(f), TesterError(FILE_NOT_FOUND % f)
        # Expected result (may be different according to flavour)
        if self.flavour:
            expectedFlavourSpecific = '%s.%s' % (expected, self.flavour)
            if os.path.exists(expectedFlavourSpecific):
                expected = expectedFlavourSpecific
        # Perform the comparison
        differ = Diff(actual, expected, areXml, xmlTagsToIgnore,
                          xmlAttrsToIgnore)
        return not differ.filesAreIdentical(report=self.report)

    def run(self):
        say = self.report.say
        say('-' * 79)
        say('- Test %s.' % self.data['Name'])
        say('- %s\n' % self.description)
        # Prepare test data
        self.tempFolder = self.testFolder / 'temp'
        if self.tempFolder.exists():
            putils.FolderDeleter.delete(self.tempFolder)
        self.tempFolder.mkdir()
        try:
            self.do()
            say('Checking result...')
            testFailed = self.checkResult()
        except:
            testFailed = self.onError()
        self.finalize()
        return testFailed

    def do(self):
        '''Concrete part of the test. Must be overridden.'''

    def checkResult(self):
        '''r_ is False if the test succeeded.'''
        return True

    def onError(self):
        '''What must happen when an exception is raised during test
           execution? Returns True if the test failed.'''
        self.errorDump = Traceback.get()
        self.report.say('Exception occurred:')
        self.report.say(self.errorDump)
        return True

    def finalize(self):
        '''Performs sme cleaning actions after test execution.'''
        pass

    def isExpectedError(self, expectedMessage):
        '''An exception was thrown. So check if the actual error message
           (stored in self.errorDump) corresponds to the p_expectedMessage.'''
        res = True
        for line in expectedMessage:
            if (self.errorDump.find(line) == -1):
                res = False
                self.report.say('"%s" not found among error dump.' % line)
                break
        return res

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TestFactory:
    @staticmethod
    def createTest(testData, testDescription, testFolder, config, flavour,
                   rendererParams):
        '''This method allows you to create tests that are instances of classes
           that you create. Those classes must be children of
           appy.shared.test.Test. m_createTest must return a Test instance and
           is called every time a test definition is encountered in the test
           plan.'''
        raise TesterError(CREATE_TEST_NOT_OVERRIDDEN)

# Program help - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
HELP_VERBOSE = 'Dumps the whole test report on stdout'
HELP_KEEP_TEMP = 'Keep the temp folder, in order to be able to copy some ' \
  'results and make them expected results when needed.'
HELP_TEST = 'Run only a specified test.'
HELP_TEST_SUITE = 'Run only a specified test suite.'
HELP_RENDERER_PARAMS = 'Specific renderer parameters, of the form ' \
  'name1=val1,name2=v2'
HELP_FLAVOUR = 'Specify the configuration flavour, which may be one of %s.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Tester(Program):
    def __init__(self, testPlan, flavours, testFactory):
        # Check test plan
        plan = Path(testPlan)
        if not plan.is_file() or not plan.name.endswith('.odt'):
            raise TesterError(WRONG_TEST_PLAN)
        self.testPlan = plan
        self.testFolder = plan.absolute().parent
        # Check flavours
        if (not isinstance(flavours,list)) and (not isinstance(flavours,tuple)):
            raise TesterError(FLAVOURS_NOT_LIST)
        for flavour in flavours:
            if not isinstance(flavour, str):
                raise TesterError(FLAVOUR_NOT_STRING)
        self.flavours = flavours
        self.flavour = None
        # Check test factory
        if not issubclass(testFactory, TestFactory):
            raise TesterError(WRONG_TEST_FACTORY)
        self.testFactory = testFactory
        # Call the base constructor: it will read and check args
        Program.__init__(self)
        # Create the report that will collect info from all executed tests
        self.report = TestReport(self.testFolder / 'Tester.report',
                                 self.verbose)

    def getRendererParams(self, params):
        '''Returns a dict with additional Renderer params'''
        r = {}
        if not params: return r
        for param in params.split(','):
            name, value = param.split('=')
            r[name] = value
        return r

    def defineArguments(self):
        '''Define the allowed arguments for this program'''
        parser = self.parser
        add = parser.add_argument
        if self.flavours:
            add('flavour', help=Make.HELP_FLAVOUR % self.flavours)
        add('-v', '--verbose', action='store_true', help=HELP_VERBOSE)
        add('-k', '--keepTemp', action='store_true', help=HELP_KEEP_TEMP)
        add('-t', '--test', help=HELP_TEST)
        add('-s', '--testSuite', help=HELP_TEST_SUITE)
        add('-r', '--rendererParams', help=HELP_RENDERER_PARAMS)

    def analyseArguments(self):
        '''Check arguments'''
        if self.flavours:
            sflavours = ', '.join(self.flavours)
            if self.flavour not in self.flavours:
                raise TesterError(WRONG_FLAVOUR % (self.flavour, sflavours))
        args = self.args
        self.verbose = args.verbose
        self.keepTemp = args.keepTemp
        self.rendererParams = self.getRendererParams(args.rendererParams)
        # If the user wants to run a single test
        self.singleTest = args.test
        # If the user wants to run a single test suite
        self.singleSuite = args.testSuite
        if self.singleTest and self.singleSuite:
            raise TesterError(TEST_AND_SUITE)

    def runSuite(self, suite):
        self.report.say('*' * 79)
        self.report.say('* Suite %s.' % suite['Name'])
        self.report.say('* %s\n' % suite['Description'])
        i = -1
        for testData in self.tables['%s.data' % suite['Name']]:
            self.nbOfTests += 1
            i += 1
            testName = testData['Name']
            if testName.startswith('_'):
                self.nbOfIgnoredTests += 1
            elif self.singleTest and (testName != self.singleTest):
                self.nbOfIgnoredTests += 1
            else:
                description = self.tables['%s.descriptions' % \
                                          suite['Name']][i]['Description']
                test = self.testFactory.createTest(
                    testData, description, self.testFolder, self.config,
                    self.flavour, self.rendererParams)
                testFailed = test.run()
                if not self.verbose:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                if testFailed:
                    self.report.say('Test failed.\n')
                else:
                    self.report.say('Test successful.\n')
                    self.nbOfSuccesses += 1
    def run(self):
        '''Runs the tests'''
        self.startTime = time.time()
        self.report.say('Parsing ODT file... ')
        self.tables = tables = TablesParser(self.testPlan).run()
        self.config = None
        ext = '.%s' % self.flavour if self.flavour else ''
        configTableName = 'Configuration%s' % ext
        if configTableName in tables:
            self.config = tables[configTableName].asDict()
        self.tempFolder = os.path.join(self.testFolder, 'temp')
        if os.path.exists(self.tempFolder):
            putils.FolderDeleter.delete(self.tempFolder)
        self.nbOfTests = 0
        self.nbOfSuccesses = 0
        self.nbOfIgnoredTests = 0
        if 'TestSuites' not in tables:
            raise TesterError(MAIN_TABLE_NOT_FOUND % str(self.testPlan))
        # Browse test suites
        for suite in tables['TestSuites']:
            if ('Name' not in suite) or ('Description' not in suite):
                raise TesterError(MAIN_TABLE_MALFORMED)
            # Must we ignore this test suite ?
            if suite['Name'].startswith('_'):
                name = suite['Name'][1:]
                tsIgnored = True
            else:
                name = suite['Name']
                tsIgnored = self.singleSuite and (name != self.singleSuite)
            # For every suite, 2 tables must exist: one with test descriptions
            # and the other with test data.
            descrTable = tables.get('%s.descriptions' % name)
            dataTable = tables.get('%s.data' % name)
            if not descrTable or not dataTable:
                raise TesterError(TEST_SUITE_NOT_FOUND % (name, name))
            if len(descrTable) != len(dataTable):
                raise TesterError(TEST_SUITE_MALFORMED % ((name,)*4))
            # The suite can still be ignored if self.singleTest is not among it
            if self.singleTest:
                inSuite = False
                for test in dataTable:
                    testName = test['Name']
                    if testName.startswith('_'): testName = testName[1:]
                    if testName == self.singleTest:
                        inSuite = True
                        break
                if not inSuite: tsIgnored = True
            if tsIgnored:
                nbOfIgnoredTests = len(dataTable)
                self.nbOfIgnoredTests += nbOfIgnoredTests
                self.nbOfTests += nbOfIgnoredTests
            else:
                self.runSuite(suite)
        self.finalize()

    def finalize(self):
        duration = time.time() - self.startTime
        msg = '%d/%d successful test(s) performed in %.2f seconds' % \
              (self.nbOfSuccesses, (self.nbOfTests-self.nbOfIgnoredTests),
               duration)
        if self.nbOfIgnoredTests >0:
            msg += ', but %d ignored test(s) not counted' % \
                   self.nbOfIgnoredTests
        msg += '.'
        self.report.say(msg, force=True)
        self.report.close()
        if not self.keepTemp:
            if os.path.exists(self.tempFolder):
                putils.FolderDeleter.delete(self.tempFolder)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy import Object
from appy.shared.utils import sequenceTypes

# ------------------------------------------------------------------------------
WRONG_LINE = 'Line number %d in file %s does not have the right number of ' \
             'fields.'

class CsvParser:
    '''This class reads a CSV file and creates a list of Python objects from it.
       The first line of the CSV file must declare the format of the following
       lines, which are 'data' lines. For example, if the first line of the file
       is

       id,roles*,password

       Then subsequent lines in the CSV need to conform to this syntax. Field
       separator will be the comma. Result of method 'parse' will be a list of
       Python objects, each one having attributes id, roles and password.
       Attributes declared with a star (like 'roles') are lists. An empty value
       will produce an empty list in the resulting object; several values need
       to be separated with the '+' sign. Here are some examples of valid 'data'
       lines for the first line above:

       gdy,,
       gdy,MeetingManager,abc
       gdy,MeetingManager+MeetingMember,abc

       In the first (and subsequent) line(s), you may choose among the following
       separators: , : ; |
       '''
    separators = [',', ':', ';', '|']
    typeLetters = {'i': int, 'f': float, 's': str, 'b': bool}
    def __init__(self, fileName, references={}, klass=None):
        self.fileName = fileName
        self.res = [] # The resulting list of Python objects.
        self.sep = None
        self.attributes = None # The list of attributes corresponding to
        # CSV columns.
        self.attributesFlags = None # Here we now if every attribute is a list
        # (True) of not (False).
        self.attributesTypes = None # Here we now the type of the attribute (if
        # the attribute is a list it denotes the type of every item in the
        # list): string, integer, float, boolean.
        self.references = references
        self.klass = klass # If a klass is given here, instead of creating
        # Object instances we will create instances of this class.
        # But be careful: we will not call the constructor of this class. We
        # will simply create instances of Object and dynamically
        # change the class of created instances to this class.

    def identifySeparator(self, line):
        '''What is the separator used in this file?'''
        maxLength = 0
        res = None
        for sep in self.separators:
            newLength = len(line.split(sep))
            if newLength > maxLength:
                maxLength = newLength
                res = sep
        self.sep = res

    def identifyAttributes(self, line):
        self.attributes = line.split(self.sep)
        self.attributesFlags = [False] * len(self.attributes)
        self.attributesTypes = [str] * len(self.attributes)
        i = -1
        for attr in self.attributes:
            i += 1
            # Is this attribute mono- or multi-valued?
            if attr.endswith('*'):
                self.attributesFlags[i] = True
            attrNoFlag = attr.strip('*')
            attrInfo = attrNoFlag.split('-')
            # What is the type of value(s) for this attribute ?
            if (len(attrInfo) == 2) and (attrInfo[1] in self.typeLetters):
                self.attributesTypes[i] = self.typeLetters[attrInfo[1]]
        # Remove trailing stars
        self.attributes = [a.strip('*').split('-')[0] for a in self.attributes]

    def resolveReference(self, attrName, refId):
        '''Finds, in self.reference, the object having p_refId.'''
        refObjects, refAttrName = self.references[attrName]
        res = None
        for refObject in refObjects:
            if getattr(refObject, refAttrName) == refId:
                res = refObject
                break
        return res

    def convertValue(self, value, basicType):
        '''Converts the atomic p_value which is a string into some other atomic
           Python type specified in p_basicType (int, float, ...).'''
        if (basicType != str) and (basicType != unicode):
            try:
                exec 'res = %s' % str(value)
            except SyntaxError, se:
                res = None
        else:   
            try:
                exec 'res = """%s"""' % str(value)
            except SyntaxError, se:
                try:
                    exec "res = '''%s'''" % str(value)
                except SyntaxError, se:
                    res = None
        return res

    def parse(self):
        '''Parses the CSV file named self.fileName and creates a list of
           corresponding Python objects (Object instances). Among object fields,
           some may be references. If it is the case, you may specify in
           p_references a dict of referred objects. The parser will then
           replace string values of some fields (which are supposed to be
           ids of referred objects) with corresponding objects in p_references.

           How does this work? p_references must be a dictionary:
           - keys correspond to field names of the current object;
           - values are 2-tuples:
             * 1st value is the list of available referred objects;
             * 2nd value is the name of the attribute on those objects that
               stores their ID.
        '''
        # The first pass parses the file and creates the Python object
        f = file(self.fileName)
        firstLine = True
        lineNb = 0
        for line in f:
            lineNb += 1
            line = line.strip()
            if not line: continue
            if firstLine:
                # The first line declares the structure of the following 'data'
                # lines.
                self.identifySeparator(line)
                self.identifyAttributes(line)
                firstLine = False
            else:
                # Add an object corresponding to this line.
                lineObject = Object()
                if self.klass:
                    lineObject.__class__ = self.klass
                i = -1
                # Do we get the right number of field values on this line ?
                attrValues = line.split(self.sep)
                if len(attrValues) != len(self.attributes):
                    raise WRONG_LINE % (lineNb, self.fileName)
                for attrValue in line.split(self.sep):
                    i += 1
                    theValue = attrValue
                    vType = self.attributesTypes[i]
                    if self.attributesFlags[i]:
                        # The attribute is multi-valued
                        if not attrValue:
                            theValue = []
                        elif '+' in theValue:
                            theValue = [self.convertValue(v, vType) \
                                        for v in attrValue.split('+')]
                        else:
                            theValue = [self.convertValue(theValue, vType)]
                    else:
                        # The attribute is mono-valued
                        theValue = self.convertValue(theValue, vType)
                    setattr(lineObject, self.attributes[i], theValue)
                self.res.append(lineObject)
        f.close()
        # The second pass resolves the p_references if any
        for attrName, refInfo in self.references.iteritems():
            if attrName in self.attributes:
                # Replace ID with real object from p_references
                for obj in self.res:
                    attrValue = getattr(obj, attrName)
                    if isinstance(attrValue, list) or \
                       isinstance(attrValue, tuple):
                        # Multiple values to resolve
                        newValue = []
                        for v in attrValue:
                            newValue.append(self.resolveReference(attrName,v))
                    else:
                        # Only one value to resolve
                        newValue = self.resolveReference(attrName, attrValue)
                    setattr(obj, attrName, newValue)
        return self.res

# ------------------------------------------------------------------------------
class CsvMarshaller:
    '''This class is responsible for producing a string, CSV-ready, line of data
       from a Appy object.'''
    undumpable = ('File', 'Action', 'Info', 'Pod')
    def __init__(self, at=None, sep=';', subSep=',', wrap='"',
                 includeHeaders=True, include=None, exclude=None):
        # If specified, p_at is an opened file handler to the CSV file to fill
        self.at = at
        # The CSV field separator
        self.sep = sep
        # The sub-separator for multi-valued fields
        self.subSep = subSep
        # The "wrap" char will wrap any value that contains self.sep.
        self.wrap = wrap
        # Must we put field names as first line in the CSV?
        self.includeHeaders = includeHeaders
        # If p_include is given, it lists names of fields that will be included
        self.include = include
        # If p_exclude is given, it lists names of fields that will be excluded
        self.exclude = exclude

    def marshallString(self, value):
        '''Produces a version of p_value that can be put in the CSV file.'''
        return value.replace('\r\n', ' ').replace('\n', ' ')

    def marshallValue(self, field, value):
        '''Produces a version of p_value that can be dumped in a CSV file.'''
        if isinstance(value, basestring):
            # Format the string as a one-line CSV-ready value
            res = self.marshallString(value)
        elif type(value) in sequenceTypes:
            # Create a list of values, separated by a sub-separator.
            res = []
            for v in value:
                res.append(self.marshallValue(field, v))
            res = self.subSep.join(res)
        elif hasattr(value, 'klass') and hasattr(value, 'title'):
            # This is a reference to another object. Dump only its title.
            res = value.title
        elif value == None:
            # Empty string is more beautiful than 'None'
            res = ''
        else:
            res = str(value)
        # If self.sep is found among this value, we must wrap it with self.wrap
        if self.sep in res:
            # Double any wrapper char if present
            res = res.replace(self.wrap, '%s%s' % (self.wrap, self.wrap))
            # Wrap the value
            res = '%s%s%s' % (self.wrap, res, self.wrap)
        return res

    def includeField(self, field):
        '''Must p_field be included in the result ?'''
        # Check self.include and self.exclude
        if self.include and field.name not in self.include: return False
        if self.exclude and field.name in self.exclude: return False
        # Check field type
        if field.type in self.undumpable: return False
        # Don't dump password fields
        if (field.type == 'String') and (field.format == 3): return False
        if (field.type == 'Ref') and field.isBack: return False
        if (field.type == 'Computed') and not field.plainText: return False
        return True

    def marshall(self, obj):
        '''Creates the CSV line representing p_obj and dumps it in self.at if
           specified, or return it else.'''
        obj = obj.appy()
        res = []
        # Dump the header line if required, and if there is still no line
        # dumped in self.at.
        headers = []
        if self.includeHeaders and self.at and (self.at.tell() == 0):
            for field in obj.fields:
                if not self.includeField(field): continue
                headers.append(field.name)
            self.at.write(self.sep.join(headers))
            self.at.write('\n')
        # Dump the data line.
        for field in obj.fields:
            if not self.includeField(field): continue
            # Get the field value
            value = field.getValue(obj.o)
            value = self.marshallValue(field, value)
            res.append(value)
        res = self.sep.join(res)
        if self.at:
            self.at.write(res)
            self.at.write('\n')
        else: return res
# ------------------------------------------------------------------------------



'''Create or update translation "po" files for an app'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import sys, collections

from appy.tr import po
from appy import Config
from appy.model.utils import Object as O
from appy.model import Config as ModelConfig

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
WRONG_APP_NAME = 'Your app is named "%s", which is not allowed.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Migrator:
    '''Migrate po(t) files from the Appy 0.9.x format to Appy 1.x format'''

    @staticmethod
    def mustMigrate(updater):
        '''Are i18n files loaded in the p_updater to the Appy 0.9.x format ?'''
        files = updater.translationFiles
        if 'Custom.pot' in files: return # There was no Custom.pot in Appy 0.9.x
        potName = '%s.pot' % updater.appName
        if potName not in files: return # There is no file at all
        pot = files[potName]
        projectName = pot.headers['Project-Id-Version'].value
        # Appy 0.9 did not insert prefix "Appy-" for this header value
        return not projectName.startswith('Appy-')

    def __init__(self, updater):
        self.updater = updater
        # Unwrap some useful updater attributes here
        self.files = updater.translationFiles
        self.appName = updater.appName
        self.model = updater.model
        # Build a dict mapping workflow prefixes in use in Appy 0.9 and
        # workflow names.
        r = {}
        for workflow in self.model.workflows.values():
            # This prefix was used, in Appy 0.9, for all i18n messages related
            # to this workflow.
            moduleName = workflow.python.__module__.split('.')[0]
            oldPrefix = '%s_%s_%s_' % (self.appName, moduleName, workflow.name)
            r[oldPrefix.lower()] = '%s_' % workflow.name
        self.workflowNames = r

    def removeSearchPart(self, id):
        '''With Appy 0.9, label parts contained some part named "search".
           With Appy 1.0, it is not the case anymore.'''
        # With Appy 1.0, searches do not have prefix "search" anymore: their
        # labels are similar to field labels.
        id = id.replace('_search_', '_')
        # With Appy 1.0, group labels are the same, be they search or field
        # groups.
        id = id.replace('_searchgroup_', '_group_')
        return id

    def migrateMessage(self, message):
        '''Converts 0.9 p_message into a 1.0 one and return this latter'''
        id = message.id
        # Check first if this m_message is related to a workflow
        for prefix, name in self.workflowNames.items():
            if id.startswith(prefix):
                id = id.replace(prefix, name)
                message.id = id
                return message
        if id.startswith('%s_' % self.appName):
            # The ID starts with a class "package name" of the form
            #               "<appName>_<moduleName>_<className>"
            # With Appy 1.x, this prefix is simply replaced with
            #                         "<className>"
            # Indeed, it is now possible because Appy 1.0 classes and workflows
            # must all have a unique name, even if put in different modules.
            id = id.split('_', 2)[2]
            id = self.removeSearchPart(id)
        else:
            # In Appy 0.9, labels for over classes start with
            #                "<appName><baseClassName>"
            # In Appy 1.0, such labels simply start with
            #                     "<baseClassName>"
            for name in self.model.baseClasses:
                prefix = '%s%s_' % (self.appName, name)
                if id.startswith(prefix):
                    id = id.replace(prefix, '%s_' % name)
                    id = self.removeSearchPart(id)
                    break
        # Set the potentially modified ID to the p_message and return it
        message.id = id
        # Replace old-fashioned carriage returns with a simple "\n" in message
        # content and default value.
        for attribute in ('msg', 'default'):
            value = getattr(message, attribute)
            if value:
                value = value.replace('<br/><br/>', '\\n')
                setattr(message, attribute, value)
        return message

    def migrateFiles(self):
        '''Performs the migration'''
        # The files to add in it after the migration
        filesToAdd = {}
        # Browse all translation files
        trFolder = self.updater.trFolder
        for name, poFile in self.files.items():
            # Extract all messages starting with ('custom_'), remove their
            # prefix and store them in a separate collection.
            customMessages = collections.OrderedDict()
            # Put any other message in "otherMessages"
            otherMessages = collections.OrderedDict()
            for id, message in poFile.messages.items():
                if id.startswith('custom_'):
                    message.id = id[7:]
                    customMessages[message.id] = message
                else:
                    message = self.migrateMessage(message)
                    otherMessages[message.id] = message
            # Add the corresponding "Custom" file in "filesToAdd"
            fileName = 'Custom.pot' if '-' not in name \
                                    else 'Custom-%s' % name.split('-', 1)[1]
            filesToAdd[fileName] = po.File(trFolder / fileName,
                                           messages=customMessages)
            # Replace poFile's message with the remaining, migrated messages
            poFile.messages = otherMessages
            # Set the "Appy 1.0" header
            poFile.headers['Project-Id-Version'].value = 'Appy-%s'% self.appName
        # Add "filesToAdd" in "files"
        for name, poFile in filesToAdd.items(): self.files[name] = poFile

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Updater:
    '''Creates or updates translation files for an app'''
    # The updater is implemented as a visitor walking a appy.model.Model
    # instance and producing the appropriate i18n labels for every model
    # element. For every such element, a method named "visit<element>" is
    # defined hereafter.

    # Default value for "confirm" labels
    CONFIRM = 'Are you sure ?'
    unallowedAppNames = ('custom', 'appy')

    def __init__(self, trFolder):
        # The "tr" sub-folder in the app. Create it if it does not exist.
        self.trFolder = trFolder
        # The app folder
        self.appFolder = trFolder.parent
        self.appName = self.appFolder.name
        if self.appName.lower() in Updater.unallowedAppNames:
            # This would clash with the translation file named Custom.pot
            raise Exception(WRONG_APP_NAME % self.appName)
        # The list of i18n labels to generate
        self.labels = collections.OrderedDict()
        # Load the model in order to get all classes, fields, workflows, states
        # and transitions for which i18n labels must be created.
        appPath = self.appFolder / '__init__.py'
        sys.path.append(str(self.appFolder.parent))
        self.config = __import__(self.appFolder.name).Config
        modelConfig = ModelConfig()
        modelConfig.set(self.appFolder)
        self.model = modelConfig.get(self.config, appOnly=True)
        # Load translation files (.po and .pot) within the "tr" folder
        self.translationFiles = po.load(trFolder)

    def addLabel(self, id, default, nice=True):
        '''Adds a label with this p_id and p_default value in self.labels'''
        self.labels[id] = po.Message(id, '', default, niceDefault=nice)

    def mustGenerateLabels(self, field):
        '''Must i18n labels be generated for p_field ?'''
        # Generating i18n labels for p_field (from a class or inner field) must
        # not be done if it uses another field's labels or if it uses custom
        # translations.
        return not field.label and not field.translations

    def visitGroup(self, class_, group, walkedGroups):
        '''Produce labels for p_group, found in p_class_ on a search or on a
           field. p_group may itself be in another group, for which labels must
           be generated, too. This recursion may continue at any level.'''
        # Do nothing if p_group is None or if labels have already been generated
        # for him.
        if not group or (group in walkedGroups): return
        if not group.label:
            # Compute the prefix used as base label for all labels to produce
            label = '%s_group_%s' % (class_.name, group.name)
            add = self.addLabel
            if group.hasLabel: add(label, group.name)
            if group.hasDescr: add('%s_descr' % label, ' ', nice=False)
            if group.hasHelp:  add('%s_help'  % label, ' ', nice=False)
            if group.hasHeaders:
                for i in range(group.nbOfHeaders):
                    add('%s_col%d' % (label, i+1), ' ', nice=False)
        # Visit group's parent
        walkedGroups.add(group)
        self.visitGroup(class_, group.group, walkedGroups)

    def visitSelect(self, class_, field, label):
        '''Generate Select-specific labels'''
        if type(field.validator) in (list, tuple):
            # The list of possible values is fixed. Generate one label for
            # every value.
            for value in field.validator:
                self.addLabel('%s_list_%s' % (label, value), value)

    def visitBoolean(self, class_, field, label):
        '''Generate Boolean-specific labels'''
        if field.render == 'radios':
            for v in ('true', 'false'):
                self.addLabel('%s_%s' % (label, v), field.yesNo[v])

    def visitAction(self, class_, field, label):
        '''Generate Action-specific labels'''
        if field.confirm:
            self.addLabel('%s_confirm' % label, self.CONFIRM, nice=False)
    visitPod = visitAction # Same labels for Pod field

    def visitRef(self, class_, field, label):
        '''Generate Ref-specific labels'''
        # Add the label for the confirm message when relevant
        if field.addConfirm:
            self.addLabel('%s_addConfirm' % label, self.CONFIRM, nice=False)

    def visitList(self, class_, field, label):
        '''Generate List-specific labels'''
        # Generate labels for sub-fields
        walkedGroups = set()
        for name, sub in field.fields:
            if not self.mustGenerateLabels(sub): continue
            # We set "forceLabel" to True because we need a label to dump for
            # every sub-field, within List' column headers.
            subLabel = '%s_%s' % (label, name)
            self.visitField(class_, sub, subLabel, walkedGroups,forceLabel=True)
    visitDict = visitList # Same labels for Dict field

    def visitCalendar(self, class_, field, label):
        '''Generate Calendar-specific labels'''
        eTypes = field.eventTypes
        if type(eTypes) not in (list, tuple): return
        for et in eTypes:
            self.addLabel('%s_event_%s' % (label, et), et)

    def visitSwitch(self, class_, field, label):
        '''Generate Switch-specific labels'''
        walkedGroups = set()
        for fieldset, fields in field.fields:
            for name, sub in fields:
                if not self.mustGenerateLabels(sub): continue
                subLabel = '%s_%s' % (class_.name, name)
                self.visitField(class_, sub, subLabel, walkedGroups,
                                forceLabel=True)

    def visitField(self, class_, field, label, walkedGroups, forceLabel=False):
        '''Generate labels for p_field. All labels will have a common prefix
           (p_label), that includes the class name and, potentially, the name of
           some outer field if p_field is an inner field.'''
        add = self.addLabel
        # Add a label for the field name when relevant
        if forceLabel or field.hasLabel or field.generateLabel:
            # Take care of sub-fields containing stars in their name
            name = field.name
            default = name.rsplit('*', 1)[-1] if '*' in name else name
            add(label, default)
        # A longer description and a help text
        if field.hasDescr: add('%s_descr' % label, ' ', nice=False)
        if field.hasHelp:  add('%s_help' % label, ' ', nice=False)
        # Generate additional labels via a field-specific "visit" method, when
        # relevant.
        method = getattr(self, 'visit%s' % field.type, None)
        if method: method(class_, field, label)
            
    def visitSearch(self, class_, search, walkedGroups):
        '''Produce labels for a p_search in a p_class_'''
        # One label for the search name and one for a longer description
        label = '%s_%s' % (class_.name, search.name)
        self.addLabel(label, search.name)
        self.addLabel('%s_descr' % label, ' ', nice=False)
        # Produce labels for the group when relevant
        self.visitGroup(class_, search.group, walkedGroups)

    def visitPhase(self, class_, phase):
        '''Produce labels for a p_phase (and inner pages) in a p_class_'''
        # Produce the label for this p_phase, excepted if it is the only phase
        # in the p_class_.
        if len(class_.phases) != 1:
            self.addLabel('%s_phase_%s' % (class_.name, phase.name), phase.name)
        # Produce one label for every page
        for page in phase.pages.values():
            if page.label is None:
                self.addLabel('%s_page_%s' % (class_.name,page.name), page.name)

    def visitClass(self, name, class_):
        '''Produce labels for a p_class_'''
        # Produce base labels for the class name (singular and plural), but
        # only for not-over classes.
        if class_.type == 'app':
            # Class name, singular and plural
            self.addLabel(name, name, nice=False)
            self.addLabel('%s_plural' % name, name + 's', nice=False)
        # While walking fields and searches, groups will be encountered. In
        # order to avoid generating labels several times for the same group,
        # remember the already walked groups.
        walkedGroups = set()
        # Produce labels for every search defined on this p_class_
        for search in class_.searches.values():
            self.visitSearch(class_, search, walkedGroups)
        # Produce labels for every field defined on this p_class_
        for field in class_.fields.values():
            # For some fields, labels must not be produced
            if self.mustGenerateLabels(field):
                label = '%s_%s' % (name, field.name)
                self.visitField(class_, field, label, walkedGroups)
            # Produce labels for the field's group(s) when relevant
            for group in field.getGroups():
                self.visitGroup(class_, group, walkedGroups)
        # Produce labels for phases and pages. If the class is "over", some
        # pages and phases are thus already defined. From the moment the user
        # adds at least one field to an existing page or phase, we consider it
        # gives him the right to rename the page and phase and we generate a
        # label for it, that will override the existing label from Appy.pot.
        for phase in class_.phases.values():
            self.visitPhase(class_, phase)

    def visitWorkflow(self, name, workflow):
        '''Produce labels for a p_class_'''
        add = self.addLabel
        # Generate a label for every state and transition
        for attribute in workflow.attributes:
            for element in getattr(workflow, attribute).values():
                label = '%s_%s' % (name, element.name)
                add(label, element.name)
                # Add a label for a transition that must be confirmed
                if (attribute == 'transitions') and element.confirm:
                    add('%s_confirm' % label, self.CONFIRM, nice=False)

    def collectLabels(self):
        '''Collect all labels by visiting model elements'''
        model = self.model
        # Add a label for every role
        for role in model.getRoles(base=None):
            self.addLabel(*role.getLabel(withDefault=False))
        # Browse model classes
        for name, class_ in model.classes.items():
            if class_.type != 'base':
                self.visitClass(name, class_)
        # Browse model workflows
        for name, workflow in model.workflows.items():
            self.visitWorkflow(name, workflow)

    def updateFiles(self):
        '''Update po(t) files based on p_self.labels. Every file that does not
           exist is created.'''
        # 2 "pot" files exist (or will exist in a few milliseconds) in
        # p_self.trFolder, with their corresponding "po" files (one per
        # supported language):
        # ----------------------------------------------------------------------
        # <appName>.pot | will contain all i18n labels automatically managed by
        #               | Appy and that were collected in p_self.labels;
        #  Custom.pot   | will contain all additional labels managed "by hand"
        #               | by the app's developer.
        # ----------------------------------------------------------------------
        counts = {True: 0, False: 0} # Count automatic and custom labels
        for pot in ('%s.pot' % self.appName, 'Custom.pot'):
            isCustom = pot == 'Custom.pot'
            # Get or create the pot file
            if pot not in self.translationFiles:
                potFile = self.trFolder / pot
                self.translationFiles[pot] = po.File(potFile)
            pot = self.translationFiles[pot]
            # Update the app's pot file with p_self.labels
            if not isCustom:
                moves = pot.update(self.labels, removeNotNew=True,
                                   keepOrder=False)
                for move in ('added', 'removed'):
                    elems = getattr(moves, move)
                    if elems:
                        print('%s - %d %s message(s): %s.' % \
                             (pot.path.name, len(elems), move, ','.join(elems)))
            pot.generate()
            # Create or update one "po" file for every language supported by the
            # app.
            labels = pot.messages if isCustom else self.labels
            counts[isCustom] = len(labels)
            for language in self.config.ui.languages:
                poName = pot.getPoFileName(language)
                if poName not in self.translationFiles:
                    poFile = self.trFolder / poName
                    self.translationFiles[poName] = po.File(poFile)
                poFile = self.translationFiles[poName]
                poFile.update(labels, removeNotNew=True, keepOrder=False)
                poFile.generate()
        # Output a summary about the operation
        total = counts[True] + counts[False]
        print('Translation files updated - %d labels (%d automatic + %d ' \
              'custom) - %d language(s).' % (total, counts[False], counts[True],
                                             len(self.config.ui.languages)))

    def run(self):
        '''Create or update translation files in p_self.appFolder'''
        # Collect, in p_self.labels, all translation labels to generate, by
        # visiting the app's model.
        self.collectLabels()
        # Create or update translation files accordingly. As a preamble, if
        # their format is still the one from Appy 0.9, migrate them.
        if Migrator.mustMigrate(self):
            Migrator(self).migrateFiles()
        self.updateFiles()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os, re, io, time, copy, collections, pathlib
from appy.model.utils import Object as O
from appy.utils.string import produceNiceMessage

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
poHeaders = '''msgid ""
msgstr ""
"Project-Id-Version: Appy-%s\\n"
"POT-Creation-Date: %s\\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=utf-8\\n"
"Content-Transfer-Encoding: 8bit\\n"
"Plural-Forms: nplurals=1; plural=0\\n"
"Language-code: %s\\n"
"Language-name: %s\\n"
"Preferred-encodings: utf-8 latin1\\n"
"Domain: %s\\n"
%s

'''
fallbacks = {'en': 'en-us en-ca',
             'fr': 'fr-be fr-ca fr-lu fr-mc fr-ch fr-fr'}

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Message:
    '''Represents a i18n message (po format)'''
    def __init__(self, id, msg, default, fuzzy=False, comments=[],
                 niceDefault=False):
        self.id = id
        self.msg = msg
        self.default = default
        if niceDefault: self.produceNiceDefault()
        # True if the default value has changed in the pot file: the msg in the
        # po file needs to be translated again.
        self.fuzzy = fuzzy 
        self.comments = comments

    def update(self, newMsg, isPot, language):
        '''Updates me with new values from p_newMsg. If p_isPot is True (which
           means that the current file is a pot file), I do not care about
           filling self.msg.'''
        if isPot:
            self.msg = ""
            if not self.default:
                self.default = newMsg.default
                # It means that if the default message has changed, we will not
                # update it in the pot file. We will always keep the one that
                # the user may have changed in the pot file. We will write a
                # default message only when no default message is defined.
        else:
            # newMsg comes from a pot file or from a base po file (like a
            # standard Appy po file). We must update the corresponding
            # message in the current po file.
            oldDefault = self.default
            if self.default != newMsg.default:
                # The default value has changed in the pot file
                oldDefault = self.default or ''
                self.default = newMsg.default
                self.fuzzy = False
                if self.msg.strip():
                    self.fuzzy = True
                    # We mark the message as "fuzzy" (=may need to be rewritten
                    # because the default value has changed) only if the user
                    # has already entered a message. Else, this has no sense to
                    # rewrite the empty message.
                    if not oldDefault.strip():
                        # This is a strange case: the old default value did not
                        # exist. Maybe was this PO file generated from some
                        # tool, but simply without any default value. So in
                        # this case, we do not consider the label as fuzzy.
                        self.fuzzy = False
            # If p_newMsg contains a message, and no message is defined for
            # self, copy it.
            if newMsg.msg and not self.msg:
                self.msg = newMsg.msg
            # For english, the the default value from a pot file can be used as
            # value for the po file.
            if (language == 'en'):
                if not self.msg:
                    # Put the default message into msg for english
                    self.msg = self.default
                if self.fuzzy and (self.msg == oldDefault):
                    # The message was equal to the old default value. It means
                    # that the user did not change it, because for English we
                    # fill by default the message with the default value (see
                    # code just above). So in this case, the message was not
                    # really fuzzy.
                    self.fuzzy = False
                    self.msg = self.default

    def produceNiceDefault(self):
        '''Transforms self.default into a nice msg'''
        self.default = produceNiceMessage(self.default)

    def generate(self):
        '''Produces myself as I must appear in a po(t) file'''
        res = ''
        for comment in self.comments:
            res += comment + '\n'
        if self.default is not None:
            res = '#. Default: "%s"\n' % self.default
        if self.fuzzy:
            res += '#, fuzzy\n'
        res += 'msgid "%s"\n' % self.id
        res += 'msgstr "%s"\n' % self.msg
        return res

    def __repr__(self):
        return '<i18n msg id="%s", msg="%s", default="%s">' % \
               (self.id, self.msg, self.default)

    def get(self):
        '''Returns self.msg, but with some replacements'''
        # Basically, the main Appy interface is the web UI. So, when formatting
        # needs to be integrated into messages as written in "po" files, we
        # expect them to be expressed in XHTML. With one exception: carriage
        # returns. Appy, for the sake of conciseness, tolerates the presence of
        # "\n" and will replace them with an XHTML carriage return "<br/>".
        # Moreover, because double quotes escaping is required due to the format
        # of "po" files, we perform such escaping too.
        return self.msg.replace('\\n', '<br/>').replace('\\"', '"')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Header:
    '''A header within a po file'''
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def generate(self):
        '''Generates the representation of myself into a po(t) file'''
        return '"%s: %s\\n"\n' % (self.name, self.value)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class File:
    '''Represents a i18n "po" file'''
    def __init__(self, path, messages=None):
        self.path = path
        self.isPot = path.suffix == '.pot'
        # Messages, ordered and keyed by their ID
        self.messages = messages or collections.OrderedDict()
        # Headers, ordered and keyed by their ID
        self.headers = collections.OrderedDict()
        # Get application name, domain name and language from fileName
        self.applicationName = ''
        self.language = ''
        self.domain = ''
        parts = path.stem.split('-')
        if self.isPot:
            if len(parts) == 1:
                self.applicationName = self.domain = parts[0]
            else:
                self.applicationName, self.domain = parts
        else:
            if len(parts) == 1:
                self.applicationName = self.domain = ''
                self.language = parts[0]
            elif len(parts) == 2:
                self.applicationName = self.domain = parts[0]
                self.language = parts[1]
            else:
                self.applicationName, self.domain, self.language = parts
        # This flag will become True during the generation process, once this
        # file will have been generated
        self.generated = False

    def __repr__(self):
        '''String representation'''
        return '<%s: %d message(s)>' % (self.path.name, len(self.messages))

    def addMessage(self, message, needsCopy=True):
        '''Adds p_message among this file's messages'''
        if needsCopy:
            message = copy.copy(message)
        self.messages[message.id] = message
        return message

    def addHeader(self, header):
        '''Adds p_message among this file's headers'''
        self.headers[header.name] = header

    def update(self, newMessages, removeNotNew=False, keepOrder=True):
        '''Updates the existing messages with p_newMessages.
           If p_removeNotNew is True, all messages in self.messages
           that are not in p_newMessages will be removed. If p_keepOrder is
           False, self.messages will be sorted according to p_newMessages. Else,
           p_newMessages that are not yet in self.messages will be appended to
           the end of self.messages.'''
        # The method returns the list of added and removed messages (IDs)
        moves = O(added=[], removed=[])
        if removeNotNew:
            for id in self.messages.keys():
                if id not in newMessages:
                    moves.removed.append(id)
        for id in moves.removed: del self.messages[id]
        # Add new messages among existing messages
        if keepOrder:
            # Update existing messages and add inexistent messages to the end
            for message in newMessages:
                if message.id in self.messages:
                    msg = self.messages[message.id]
                else:
                    msg = self.addMessage(message)
                    moves.added.append(message.id)
                msg.update(message, self.isPot, self.language)
        else:
            # Remove, from self.messages, all messages not being in new
            # messages. We will append them at the end of the new messages.
            if not removeNotNew:
                notNew = [msg for msg in self.messages.values() \
                          if msg.id not in newMessages]
                for msg in notNew: del self.messages[msg.id]
            # Keep existing messages and start with a fresh dict
            existing = self.messages
            self.messages = collections.OrderedDict()
            for message in newMessages.values():
                if message.id in existing:
                    msg = existing[message.id]
                    self.messages[message.id] = msg
                else:
                    msg = self.addMessage(message)
                    moves.added.append(message.id)
                msg.update(message, self.isPot, self.language)
            # Append the list of old messages to the end
            if not removeNotNew:
                for message in notNew: self.messages[message.id] = message
        return moves

    def generateHeaders(self, f):
        if not self.headers:
            creationTime = time.strftime("%Y-%m-%d %H:%M-%S", time.localtime())
            fb = ''
            if not self.isPot:
                # I must add fallbacks
                if self.language in fallbacks:
                    fb = '"X-is-fallback-for: %s\\n"' % fallbacks[self.language]
            f.write(poHeaders % (self.applicationName, creationTime,
                                 self.language, self.language, self.domain, fb))
        else:
            # Some headers were already found, we dump them as is
            f.write('msgid ""\nmsgstr ""\n')
            for header in self.headers.values():
                f.write(header.generate())
            f.write('\n')

    def generate(self, inFile=True):
        '''Generates the corresponding po(t) file on disk if p_inFile is True,
           in a io.StringIO else. In this latter case, the StringIO instance is
           returned.'''
        if inFile:
            f = open(str(self.path), 'w')
        else:
            f = io.StringIO()
        self.generateHeaders(f)
        for msg in self.messages.values():
            f.write(msg.generate())
            f.write('\n')
        if inFile:
            f.close()
        self.generated = True
        return f

    def getPoFileName(self, language):
        '''Gets the name of the po file that corresponds to this pot file and
           the given p_language.'''
        if self.applicationName == self.domain:
            r = '%s-%s.po' % (self.applicationName, language)
        else:
            r = '%s-%s-%s.po' % (self.applicationName, self.domain, language)
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Parser:
    '''Parse a translation file in po(t) format. The result is produced in
       self.res as a File instance.'''
    def __init__(self, path):
        self.res = File(path)

    # Regular expressions for msgIds, msgStrs and default values.
    rexDefault = re.compile('#\.\s+Default\s*:\s*"(.*)"')
    rexFuzzy = re.compile('#,\s+fuzzy')
    rexId = re.compile('msgid\s+"(.*)"')
    rexMsg = re.compile('msgstr\s+"(.*)"')

    def parse(self):
        '''Parse all i18n messages from the file and return the corresponding
           File instance.'''
        f = open(str(self.res.path), encoding='utf-8')
        # Currently parsed values
        msgDefault = msgFuzzy = msgId = msgStr = None
        comments = []
        # Walk every line of the po(t) file
        for line in f:
            lineContent = line.strip()
            if lineContent and (not lineContent.startswith('"')):
                r = self.rexDefault.match(lineContent)
                if r:
                    msgDefault = r.group(1)
                else:
                    r = self.rexFuzzy.match(lineContent)
                    if r:
                        msgFuzzy = True
                    else:
                        r = self.rexId.match(lineContent)
                        if r:
                            msgId = r.group(1)
                        else:
                            r = self.rexMsg.match(lineContent)
                            if r:
                                msgStr = r.group(1)
                            else:
                                if lineContent.startswith('#'):
                                    comments.append(lineContent.strip())
                if msgStr is not None:
                    if not ((msgId == '') and (msgStr == '')):
                        poMsg = Message(msgId, msgStr, msgDefault, msgFuzzy,
                                        comments)
                        self.res.addMessage(poMsg, needsCopy=False)
                    msgDefault = msgFuzzy = msgId = msgStr = None
                    comments = []
            if lineContent.startswith('"'):
                # It is a header value
                name, value = lineContent.strip('"').split(':', 1)
                if value.endswith('\\n'):
                    value = value[:-2]
                self.res.addHeader(Header(name.strip(), value.strip()))
        f.close()
        return self.res

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def load(path=None, pot=True, languages=None):
    '''Loads all "po" and "pot" files (or only "po" files if p_pot is False)
       wihin p_path, or within appy/tr/po if p_path is None. If p_languages is
       not None, we only read "po[t]" files related to these languages.'''
    r = {} # ~{s_fileName: File}~
    # Determine the path where to find po[t] files: the app's "tr" sub-folder or
    # Appy's "tr/po" sub-folder.
    path = path or pathlib.Path(__file__).parent
    # Search "po" files only, or "po" and "pot" files
    suffix = '*' if pot else ''
    # Browse files
    for poFile in path.glob('*.po%s' % suffix):
        # Dismiss temp files, ie .po.swp
        if poFile.suffix not in ('.po', '.pot'): continue
        if languages:
            # Bypass this file if its language is not among p_languages
            stem = poFile.stem
            language = stem if '-' not in stem else stem.rsplit('-', 1)[-1]
            if language not in languages: continue
        parser = Parser(poFile)
        r[poFile.name] = parser.parse()
    return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import inspect
from persistent import Persistent

from appy.px import Px
from appy.ui.js import Quote
from appy.model.base import Base
from appy.model.fields import Field
from appy.server.request import Request
from appy.model.workflow.transition import Transition

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
UNTRAVERSABLE = 'Untraversable element "%s" while traversing %s.'
TRAVERSE_UNKNOWN = 'Unknown traversal type "%s".'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class NotFound(Exception):
    '''Exception raised when an URL path does not correspond to an existing (or
       visible) object.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Traversal:
    '''Traverses the path of an URL and executes the corresponding code'''
    Error = NotFound

    # Let's define an URL as
    #               <protocol>://<domain name><path>
    # "Traversing" is the process of splitting <path> into parts and finding a
    # correpondance between these parts and database objects, their attributes
    # and methods. Here are some examples.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # <path> content | Description
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # /              | The root URL
    # /tool/view     | Attribute "view" on object having id "tool"
    # /a/b/c/d?var=1 | Evaluating attribute or method "b" on object "a" produces
    #                | a partial result "r". Then, evaluating "c" on "r"
    #                | produces "r2". Evaluating "d" or "r2" produces the final
    #                | result. While traversing all these parts, request
    #                | variable named "var" holding value "1", is available.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # "Traversing" thus finally produces a result as a page that is returned to
    # the browser.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # When the traversal is in the default, "standard" mode, fields are supposed
    # to be related to the currently walked object. If the traversal switches to
    # "static" mode, fields are supposed to run independently of any instance.
    # A switch from standard to static mode occurs as soon as a @-prefixed part
    # is encountered.
    STANDARD = 0
    STATIC   = 1

    def __init__(self, handler=None, other=None):
        # A traversal can be initialised from another one
        if other:
            self.init(other)
            self.parent = other
            return
        # Perform standard initialisation
        self.handler = handler
        self.req = handler.req
        # The path that will be traversed, splitted into parts
        self.parts = handler.parts
        # The currently logged user
        self.user = handler.guard.user
        # The current result of traversing
        self.r = None
        # The last encountered object
        self.o = None
        # The last encountered field
        self.field = None
        # The last encountered transition
        self.transition = None
        # The PX context
        self.context = None
        # The parent traversal, for a child traversal
        self.parent = None
        # The traversal mode
        self.mode = Traversal.STANDARD

    def init(self, other):
        for name, value in other.__dict__.items():
            setattr(self, name, value)

    def getPath(self):
        '''Returns the complete path, including, for a child traversal, the
           parent path.'''
        r = '/'.join(self.parts)
        if self.parent:
            r += ' (from %s)' % self.parent.getPath()
        return r

    def createContext(self, layout='view'):
        '''Create and return a PX context for rendering p_o on p_layout'''
        return Px.createContext(self, layout, Quote)

    def asObjectId(self, part):
        '''Interpret p_part as an object ID and return this ID as
           * a positive integer (for a definitive object having a standard ID);
           * a negative integer (for a temporary object);
           * p_part itself if it does not represent a (positive or negative)
             integer: it is assumed to represent a custom string ID for a
             definitive object.
        '''
        if part.isdigit() or (part.startswith('-') and part[1:].isdigit()):
            return int(part)
        return part

    def getObject(self, part):
        '''Tries to get an object from "part" which is a part of the URL path'''
        if part:
            id = self.asObjectId(part)
            r = self.handler.server.database.getObject(self.handler, id)
            if not r: raise NotFound(self.getPath())
        else:
            # If "part" is empty, (path is "/"), use the tool as default object
            r = self.handler.tool
        # Check if the user is allowed to consult this object
        self.handler.guard.allows(r, raiseError=True)
        return r

    def getSpecialObject(self, name):
        '''Gets the special object named p_name'''
        # Currenly, the only type of supported "special object" is an app class
        return self.handler.server.model.classes.get(name).python

    def getLayout(self, pxName):
        '''Determine the layout type from the currently traversed PX'''
        return 'view' if pxName == 'default' else pxName

    def getPermission(self, permission):
        '''If a field is currently traversed, a "read" or "write" permission
           must be converted to this field's potentially specific "read" or
           "write" permission.'''
        field = self.field
        if not field: return permission
        if permission == 'read':
            r = field.readPermission
        elif permission == 'write':
            r = field.writePermission
        else:
            r = permission
        return r

    def allowTraversal(self, info, o, name, raiseError=True):
        '''Raise an exception if the user is not allowed to traverse this
           p_o(bject). Traversal info is given in p_info.'''
        if isinstance(info, bool):
            r = info
        elif isinstance(info, str):
            if ':' not in info:
                # This is a role: the user must have it
                r = self.user.hasRole(info)
            else:
                type, info = info.split(':', maxsplit=1)
                if type == 'perm':
                    # The user must have this permission on p_o
                    permission = self.getPermission(info)
                    r = o.allows(permission=permission)
                elif type == 'user':
                    # The user must be the one whose login is in "info"
                    r = self.user.login == info
                else:
                    raise self.handler.guard.Error(TRAVERSE_UNKNOWN % type)
        else:
            # A tuple or list of logically OR'ed infos. Having at least one of
            # the listed prerogatives allows the user to traverse.
            r = False
            for inf in info:
                if self.allowTraversal(inf, o, name, raiseError=False):
                    r = True
                    break
        # Return the result or raise an Unauthorized exception if p_raiseError
        # is True and the object can't be traversed.
        if raiseError and not r:
            message = UNTRAVERSABLE % (name, self.getPath())
            raise self.handler.guard.Error(message)
        return r

    def getTraversalInfo(self, previous, name, current):
        '''Get traversal info about p_previous.<p_name> = p_current'''
        # Do we have traversal information ? p_previous can be an object, a
        # field, a module...
        traverse = getattr(previous, 'traverse', None)
        # Is p_name mentioned in the traversal info ?
        if traverse and (name in traverse):
            # Yes: retrieve "traverse info" from it
            r = traverse[name]
        else:
            # No: determine default "traverse info" based on p_current's type
            if isinstance(current, Field):
                if self.mode == Traversal.STANDARD:
                    # One may traverse it if he has the read permission on the
                    # related object.
                    r = 'perm:%s' % current.readPermission
                else:
                    # The field is not related to any instance. At this
                    # traversal level, no more security check can be performed.
                    r = True
            elif isinstance(current, Transition):
                # The transition is traversable if triggerable
                r = current.isTriggerable(previous)
            elif isinstance(current, Px):
                # A PX is traversable by default
                r = True
            else:
                # Any other element (method, attribute, property...) is not
                # traversable by default.
                r = False
        # If we are walking an element of a Field in static mode, and this
        # element is protected by an instance-related info (=a permission),
        # convert this info to True: indeed, in that case, permission can't be
        # checked (we have no instance).
        if (self.mode == Traversal.STATIC) and isinstance(previous, Field) and \
           isinstance(r, str) and r.startswith('perm:'):
            r = True
        return r

    def executeMethod(self, method):
        '''Execute m_method and return the result. The args potentially given to
           the method may vary.'''
        if hasattr(method, '__self__'):
            # We have a bound method. If the last walked object is not the bound
            # object, give the last walked object as unique arg.
            o = self.o
            if o and (method.__self__ != o):
                r = method(o)
            else:
                r = method()
        else:
            r = method(self.o or self.handler)
        return r

    def executePx(self, name, px):
        '''Renders this p_px, whose p_name is given'''
        # Firstly, create or update the PX context
        layout = self.getLayout(name)
        if not self.context:
            self.context = self.createContext(layout)
        else:
            # Update the existing context
            self.context.o = self.o
            self.context.field = self.field
            self.context.layout = layout
        # Secondly, execute the tied action if defined
        actionName = self.req.action
        if actionName:
            # Execute this action by performing a sub-traversal
            handler = self.handler
            sub = Traversal(other=self)
            sub.r = sub.field or sub.o
            sub.parts = actionName.split('*')
            # A database commit will probably be required
            handler.commit = True
            # Run the sub-traversal
            msg = sub.run()
            if msg:
                # Set a special cookie in the response containing the message
                handler.resp.addMessage(msg)
        # Finally, call the PX
        return px(self.context)

    def managePart(self, previous, name, current):
        '''Manage the currently traversed element p_current, named p_name, after
           p_previous has already been traversed.

                       p_previous.<p_name> = p_current
        '''
        # Compute information allowing to determine if the user is allowed to
        # perform such traversal.
        info = self.getTraversalInfo(previous, name, current)
        # Raise an Unauthorized exception if the user is not allowed to do it
        self.allowTraversal(info, self.o, name)
        # "Execute" the p_current object
        if isinstance(current, Px):
            # p_current is a PX: render it in p_self.r
            self.r = self.executePx(name, current)
        elif callable(current) and not inspect.isclass(current):
            # Execute method "current"
            self.r = self.executeMethod(current)
        else:
            # Ensure p_current is already in p_self.r. This may not be the case
            # at the end of the traversal, if we add a default value.
            self.r = current

    def getPart(self, part):
        '''Get the attribute or method corresponding to p_part on the current
           result in p_self.r.'''
        # If p_self.r is an object and p_part corresponds to a specific
        # sub-element on this object (field or transition), return this
        # sub-element.
        r = self.r
        if isinstance(r, Base):
            # Is p_part a field defined on the current object ?
            field = r.getField(part)
            if field:
                self.field = field
                return field
            # Is p_part a transition defined on the current object's workflow ?
            transition = r.getWorkflow().transitions.get(part)
            if transition:
                self.transition = transition
                return transition
        # Try to get an attribute or method on p_self.r named p_part
        r = getattr(r, part, None)
        if r is None:
            # Raise a 404
            raise NotFound(self.getPath())
        if isinstance(r, Base):
            self.o = r
            self.field = self.transition = None
        return r

    def run(self):
        '''Traverses a given path whose parts are splitted into p_self.parts'''
        parts = self.parts
        if not parts:
            # There is nothing to traverse. Define the tool as the default
            # object to traverse.
            self.r = self.o = self.getObject(None)
        else:
            # Traverse URL parts
            length = len(parts)
            i = 0
            # Define the previously traversed part. It is None for a main
            # traversal, or an already filled self.r in the case of a
            # sub-traversal.
            previous = self.r
            while i < length:
                # Get and manage the next part
                part = parts[i]
                if part and part.startswith('@'):
                    # Manage direct navigation to a special object
                    self.r = self.getSpecialObject(part[1:])
                    # Switch to "static" mode
                    self.mode = Traversal.STATIC
                elif self.r is None:
                    # If we are at the start of the traversal, and "r" is still
                    # None, "part" must correspond to an object ID.
                    self.r = self.o = self.getObject(part)
                # Manage any not-empty part
                elif part:
                    # Try to get the attribute or method named "part" on self.r
                    self.r = self.getPart(part)
                    # Manage this part
                    self.managePart(previous, part, self.r)
                i += 1
                previous = self.r
        # If self.r is an object, call the default PX on it
        if isinstance(self.r, Persistent):
            self.managePart(self.r, 'default', self.r.default)
        # If the result of the traversal is not a string, it is invalid. Produce
        # a 404 error in that case.
        r = self.r
        if r is not None and not isinstance(r, str):
            raise NotFound(self.getPath())
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This script allows to check a LDAP connection.'''
import sys
from appy.shared.ldap_connector import LdapConnector

# ------------------------------------------------------------------------------
class LdapTester:
    '''Usage: python checkldap.py ldapUri login password base attrs filter scope

         ldapUri is, for example, "ldap://127.0.0.1:389"
         login is the login user DN, ie: "cn=gdy,o=geezteem"
         password is the password for this login
         base is the base DN where to perform the search, ie "ou=hr,o=GeezTeem"
         attrs is a comma-separated list of attrs we will retrieve in the LDAP,
               ie "uid,login"
         filter is the query filter, ie "(&(attr1=Geez*)(status=OK))"
         scope is the scope of the search, and can be:
           BASE     To search the object itself on base
           ONELEVEL To search base's immediate children
           SUBTREE  To search base and all its descendants
    '''
    def __init__(self):
        # Get params from shell args.
        if len(sys.argv) != 8:
            print(LdapTester.__doc__)
            sys.exit(0)
        s = self
        s.uri,s.login,s.password,s.base,s.attrs,s.filter,s.scope = sys.argv[1:]
        self.attrs = self.attrs.split(',')
        self.tentatives = 5
        self.timeout = 5
        self.attributes = ['cn']
        self.ssl = False

    def test(self):
        # Connect the the LDAP
        print('Connecting to... %s' % self.uri)
        connector = LdapConnector(self.uri)
        success, msg = connector.connect(self.login, self.password)
        if not success: return
        # Perform the query.
        print ('Querying %s...' % self.base)
        res = connector.search(self.base, self.scope, self.filter,
                               self.attributes)
        print('Got %d results' % len(res))

# ------------------------------------------------------------------------------
if __name__ == '__main__': LdapTester().test()
# ------------------------------------------------------------------------------
